from ..environment import torch, math

'''
1.class NWKernelRegression(torch.nn.Module)
summary:基于带参数的注意力汇聚，使用小批量矩阵乘法， 定义Nadaraya-Watson核回归的带参数版本为
'''
class NWKernelRegression(torch.nn.Module):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.w = torch.nn.Parameter(torch.rand((1,), requires_grad=True))

    def forward(self, queries, keys, values):
        # queries和attention_weights的形状为(查询个数，“键－值”对个数)
        queries = queries.repeat_interleave(keys.shape[1]).reshape((-1, keys.shape[1]))
        self.attention_weights = torch.nn.functional.softmax(
            -((queries - keys) * self.w)**2 / 2, dim=1)
        # values的形状为(查询个数，“键－值”对个数)
        return torch.bmm(self.attention_weights.unsqueeze(1),
                         values.unsqueeze(-1)).reshape(-1)


'''
2.class AdditiveAttention(torch.nn.Module)
    加性注意力
'''
class AdditiveAttention(torch.nn.Module):
    '''加性注意力'''
    def __init__(self, key_size, query_size, num_hiddens, dropout, **kwargs):
        super(AdditiveAttention, self).__init__(**kwargs)
        self.W_k = torch.nn.Linear(key_size, num_hiddens, bias=False)
        self.W_q = torch.nn.Linear(query_size, num_hiddens, bias=False)
        self.W_v = torch.nn.Linear(num_hiddens,1,bias=False)
        self.dropout = torch.nn.Dropout(dropout)

    def forward(self, queries, keys, values, valid_lens):
        from ..datasets.dataset_utils import masked_softmax


        queries, keys = self.W_q(queries), self.W_k(keys)
        # 在维度扩展后，
        # queries的形状：(batch_size，查询的个数，1，num_hidden)
        # key的形状：(batch_size，1，“键－值”对的个数，num_hiddens)
        # 使用广播方式进行求和
        features = queries.unsqueeze(2) + keys.unsqueeze(1)
        features = torch.tanh(features)
        # self.w_v仅有一个输出，因此从形状中移除最后那个维度。
        # scores的形状：(batch_size，查询的个数，“键-值”对的个数)
        scores = self.W_v(features).squeeze(-1)
        self.attention_weights = masked_softmax(scores, valid_lens)
        # values的形状：(batch_size，“键－值”对的个数，值的维度)
        return torch.bmm(self.dropout(self.attention_weights), values)



'''
3.class DotProductAttention(torch.nn.Module)
    缩放点积注意力
'''
#@save
class DotProductAttention(torch.nn.Module):
    """缩放点积注意力"""
    def __init__(self, dropout, **kwargs):
        super(DotProductAttention, self).__init__(**kwargs)
        self.dropout = torch.nn.Dropout(dropout)

    # queries的形状：(batch_size，查询的个数，d)
    # keys的形状：(batch_size，“键－值”对的个数，d)
    # values的形状：(batch_size，“键－值”对的个数，值的维度)
    # valid_lens的形状:(batch_size，)或者(batch_size，查询的个数)
    def forward(self, queries, keys, values, valid_lens=None):
        from ..datasets.dataset_utils import masked_softmax
        d = queries.shape[-1]
        # 设置transpose_b=True为了交换keys的最后两个维度
        scores = torch.bmm(queries, keys.transpose(1,2)) / math.sqrt(d)
        self.attention_weights = masked_softmax(scores, valid_lens)
        return torch.bmm(self.dropout(self.attention_weights), values)





'''
4.class MultiHeadAttention(torch.nn.Module)
'''
class MultiHeadAttention(torch.nn.Module):
    '''多头注意力'''

    def __init__(self, key_size, query_size, value_size, num_hiddens, num_heads, dropout, bias=False, **kwargs):
        super(MultiHeadAttention, self).__init__()
        self.num_heads = num_heads
        self.attention = DotProductAttention(dropout)
        self.W_q = torch.nn.Linear(query_size, num_hiddens, bias=bias)
        self.W_k = torch.nn.Linear(key_size, num_hiddens, bias=bias)
        self.W_v = torch.nn.Linear(value_size, num_hiddens, bias=bias)
        self.W_o = torch.nn.Linear(num_hiddens, num_hiddens, bias=bias)

    def forward(self, queries, keys, values, valid_lens):
        # queries, keys, values的形状
        # (batch_size，查询或者“键－值”对的个数，num_hiddens)
        # valid_lens 的形状
        # (batch_size，)或(batch_size，查询的个数)
        # 经过变换后，输出的queries，keys，values　的形状:
        # (batch_size*num_heads，查询或者“键－值”对的个数，
        # num_hiddens/num_heads)
        queries = self.transpose_qkv(self.W_q(queries), self.num_heads)
        keys = self.transpose_qkv(keys, self.num_heads)
        values = self.transpose_qkv(values, self.num_heads)

        if valid_lens is not None:
            # 在轴0，将第一项（标量或者矢量）复制num_heads次，
            # 然后如此复制第二项，然后诸如此类。
            valid_lens = torch.repeat_interleave(valid_lens, repeats=self.num_heads, dim=0)

        # output的形状:(batch_size*num_heads，查询的个数，
        # num_hiddens/num_heads)
        output = self.attention(queries, keys, values, valid_lens)

        # output_concat的形状:(batch_size，查询的个数，num_hiddens)
        output_concat = self.transpose_output(output, self.num_heads)
        return self.W_o(output_concat)

    def transpose_qkv(self, X, num_heads):
        """为了多注意力头的并行计算而变换形状"""
        # 输入X的形状:(batch_size，查询或者“键－值”对的个数，num_hiddens)
        # 输出X的形状:(batch_size，查询或者“键－值”对的个数，num_heads，
        # num_hiddens/num_heads)
        X = X.reshape(X.shape[0], X.shape[1], num_heads, -1)

        # 输出X的形状:(batch_size，num_heads，查询或者“键－值”对的个数,
        # num_hiddens/num_heads)
        X = X.permute(0, 2, 1, 3)

        # 最终输出的形状:(batch_size*num_heads,查询或者“键－值”对的个数,
        # num_hiddens/num_heads)
        return X.reshape(-1, X.shape[2], X.shape[3])

    def transpose_output(self, X, num_heads):
        """逆转transpose_qkv函数的操作"""
        X = X.reshape(-1, num_heads, X.shape[1], X.shape[2])
        X = X.permute(0, 2, 1, 3)
        return X.reshape(X.shape[0], X.shape[1], -1)






from ..environment import torch, math, np
from ..environment import torchvision



'''
img_shape:指图像形状，[ch, height, width]
'''


class Generator(torch.nn.Module):
    '''
    Generator model of GAN
    '''
    def __init__(self, latent_dim, output_shape):
        super(Generator, self).__init__()
        self.output_shape = output_shape


        def block(in_feat, out_feat, normalize=True):
            layers = [torch.nn.Linear(in_feat, out_feat)]
            if normalize:
                layers.append(torch.nn.BatchNorm1d(out_feat, 0.8))
            layers.append(torch.nn.LeakyReLU(0.2, inplace=True))
            return layers

        self.model = torch.nn.Sequential(
            *block(latent_dim, 128, normalize=False),
            *block(128, 256),
            *block(256, 512),
            *block(512, 1024),
            torch.nn.Linear(1024, int(np.prod(output_shape))),
            torch.nn.Tanh()
        )

    def forward(self, z):
        img = self.model(z)
        img = img.view(img.size(0), *self.output_shape)
        return img







class Discriminator(torch.nn.Module):
    def __init__(self, img_shape):
        super(Discriminator, self).__init__()

        self.model = torch.nn.Sequential(
            torch.nn.Linear(int(np.prod(img_shape)), 512),
            torch.nn.LeakyReLU(0.2, inplace=True),
            torch.nn.Linear(512, 256),
            torch.nn.LeakyReLU(0.2, inplace=True),
            torch.nn.Linear(256, 1),
            torch.nn.Sigmoid(),
        )

    def forward(self, img):
        img_flat = img.view(img.size(0), -1)
        validity = self.model(img_flat)

        return validity



from PepperPepper.environment import torch, nn, einops, rearrange, Rearrange

# 空间注意力
class SpatialAttention(nn.Module):
    def __init__(self):
        super(SpatialAttention, self).__init__()
        self.sa = nn.Conv2d(2, 1, 7, padding=3, padding_mode='reflect', bias=True)

    def forward(self, x):
        x_avg = torch.mean(x, dim=1, keepdim=True)
        x_max, _ = torch.max(x, dim=1, keepdim=True)
        x2 = torch.cat([x_avg, x_max], dim=1)
        sattn = self.sa(x2)
        return sattn

class ChannelAttention(nn.Module):
    def __init__(self, dim, reduction=3):
        super(ChannelAttention, self).__init__()
        self.gap = nn.AdaptiveAvgPool2d(1)
        self.ca = nn.Sequential(
            nn.Conv2d(dim, dim // reduction, 1, padding=0, bias=True),
            nn.LeakyReLU(inplace=True),
            nn.Conv2d(dim // reduction, dim, 1, padding=0, bias=True),
        )

    def forward(self, x):
        x_gap = self.gap(x)
        cattn = self.ca(x_gap)
        return cattn



class PixelAttention(nn.Module):
    def __init__(self, dim):
        super(PixelAttention, self).__init__()
        self.pa2 = nn.Conv2d(2 * dim, dim, 7, padding=3, padding_mode='reflect', groups=dim, bias=True)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x, pattn1):
        B, C, H, W = x.shape
        x = x.unsqueeze(dim=2)  # B, C, 1, H, W
        pattn1 = pattn1.unsqueeze(dim=2)  # B, C, 1, H, W
        x2 = torch.cat([x, pattn1], dim=2)  # B, C, 2, H, W
        x2 = Rearrange('b c t h w -> b (c t) h w')(x2)
        pattn2 = self.pa2(x2)
        pattn2 = self.sigmoid(pattn2)
        return pattn2







# sobel注意力机制
class SobelAttention(torch.nn.Module):
    def __init__(self, in_channels):
        super(SobelAttention, self).__init__()
        self.channels = in_channels

        # sobel算子
        self.sobel_v = torch.nn.Conv2d(in_channels=in_channels, out_channels=in_channels, kernel_size=3, padding=1,
                                       bias=False)
        self.sobel_h = torch.nn.Conv2d(in_channels=in_channels, out_channels=in_channels, kernel_size=3, padding=1,
                                       bias=False)

        # 初始化卷积核
        sobel_kernel_v = torch.tensor([[0, -1, 0],
                                       [0, 0, 0],
                                       [0, 1, 0]], dtype=torch.float32).view(1, 1, 3, 3)
        sobel_kernel_h = torch.tensor([[0, 0, 0],
                                       [-1, 0, 1],
                                       [0, 0, 0]], dtype=torch.float32).view(1, 1, 3, 3)

        # 将卷积核扩展到多个通道
        sobel_kernel_v = sobel_kernel_v.repeat(in_channels, in_channels, 1, 1)
        sobel_kernel_h = sobel_kernel_h.repeat(in_channels, in_channels, 1, 1)

        self.sobel_v.weight = torch.nn.Parameter(sobel_kernel_v, requires_grad=False)
        self.sobel_h.weight = torch.nn.Parameter(sobel_kernel_h, requires_grad=False)



        # sobel算子
        self.sobel3_x = torch.nn.Conv2d(in_channels=in_channels, out_channels=in_channels, kernel_size=3, padding=1, bias=False, padding_mode='reflect')
        self.sobel3_y = torch.nn.Conv2d(in_channels=in_channels, out_channels=in_channels, kernel_size=3, padding=1, bias=False, padding_mode='reflect')
        self.sobel5_x = torch.nn.Conv2d(in_channels=in_channels, out_channels=in_channels, kernel_size=5, padding=2, bias=False, padding_mode='reflect')
        self.sobel5_y = torch.nn.Conv2d(in_channels=in_channels, out_channels=in_channels, kernel_size=5, padding=2, bias=False, padding_mode='reflect')
        self.sobel7_x = torch.nn.Conv2d(in_channels=in_channels, out_channels=in_channels, kernel_size=7, padding=3, bias=False, padding_mode='reflect')
        self.sobel7_y = torch.nn.Conv2d(in_channels=in_channels, out_channels=in_channels, kernel_size=7, padding=3, bias=False, padding_mode='reflect')

        # 初始化卷积核
        sobel_kernel3_x = torch.tensor([[1, 0, -1],
                                       [2, 0, -2],
                                       [1, 0, -1]], dtype=torch.float32).view(1, 1, 3, 3)
        sobel_kernel3_y = torch.tensor([[1, 2, 1],
                                       [0, 0, 0],
                                       [-1, -2, -1]], dtype=torch.float32).view(1, 1, 3, 3)

        sobel_kernel5_x = torch.tensor(
            [[-2, -1, 0, 1, 2],
             [-4, -3, 0, 3, 4],
             [-6, -5, 0, 5, 6],
             [-4, -3, 0, 3, 4],
             [-2, -1, 0, 1, 2]], dtype=torch.float32
        ).view(1,1,5,5)

        sobel_kernel5_y = torch.tensor(
            [[2, 4, 6, 4, 2],
             [1, 3, 5, 3, 1],
             [0, 0, 0, 0, 0],
             [-1, -3, -5, -3, -1],
             [-2, -4, -6, -4, -2]], dtype=torch.float32
        ).view(1,1,5,5)

        sobel_kernel7_x = torch.tensor(
            [[-3, -2, -1, 0, 1, 2, 3],
             [-6, -5, -4, 0, 4, 5, 6],
             [-9, -8, -7, 0, 7, 8, 9],
             [-12, -10, -9, 0, 9, 10, 12],
             [-9, -8, -7, 0, 7, 8, 9],
             [-6, -5, -4, 0, 4, 5, 6],
             [-3, -2, -1, 0, 1, 2, 3]], dtype=torch.float32
        ).view(1,1,7,7)

        sobel_kernel7_y = torch.tensor(
            [[3, 6, 9, 12, 9, 6, 3],
             [2, 5, 8, 10, 8, 5, 2],
             [1, 4, 7, 9, 7, 4, 1],
             [0, 0, 0, 0, 0, 0, 0],
             [-1, -4, -7, -9, -7, -4, -1],
             [-2, -5, -8, -10, -8, -5, -2],
             [-3, -6, -9, -12, -9, -6, -3]], dtype=torch.float32
        ).view(1,1,7,7)


        # 将卷积核扩展到多个通道
        sobel_kernel3_x = sobel_kernel3_x.repeat(in_channels, in_channels, 1, 1)
        sobel_kernel3_y = sobel_kernel3_y.repeat(in_channels, in_channels, 1, 1)

        sobel_kernel5_x = sobel_kernel5_x.repeat(in_channels, in_channels, 1, 1)
        sobel_kernel7_x = sobel_kernel7_x.repeat(in_channels, in_channels, 1, 1)
        sobel_kernel5_y = sobel_kernel5_y.repeat(in_channels, in_channels, 1, 1)
        sobel_kernel7_y = sobel_kernel7_y.repeat(in_channels, in_channels, 1, 1)

        self.sobel3_x.weight = torch.nn.Parameter(sobel_kernel3_x, requires_grad=False)
        self.sobel3_y.weight = torch.nn.Parameter(sobel_kernel3_y, requires_grad=False)
        self.sobel5_x.weight = torch.nn.Parameter(sobel_kernel5_x, requires_grad=False)
        self.sobel5_y.weight = torch.nn.Parameter(sobel_kernel5_y, requires_grad=False)
        self.sobel7_x.weight = torch.nn.Parameter(sobel_kernel7_x, requires_grad=False)
        self.sobel7_y.weight = torch.nn.Parameter(sobel_kernel7_y, requires_grad=False)

        self.sobelbn = torch.nn.InstanceNorm2d(in_channels * 3)
        self.sobelbn3 = torch.nn.InstanceNorm2d(in_channels * 3)
        self.sobelbn5 = torch.nn.InstanceNorm2d(in_channels * 3)
        self.sobelbn7 = torch.nn.InstanceNorm2d(in_channels * 3)

        self.CA = ChannelAttention(in_channels * 12)
        self.SA = SpatialAttention()
        self.PA = PixelAttention(in_channels * 12)



        # 定义线性变换
        self.conv33 = torch.nn.Conv2d(in_channels=in_channels * 12, out_channels=in_channels, kernel_size=3, padding=1, bias=True)
        self.bn3 = torch.nn.InstanceNorm2d(in_channels)
        self.conv55 = torch.nn.Conv2d(in_channels=in_channels * 12, out_channels=in_channels, kernel_size=5, padding=2, bias=True)
        self.bn5 = torch.nn.InstanceNorm2d(in_channels)
        self.conv77 = torch.nn.Conv2d(in_channels=in_channels * 12, out_channels=in_channels, kernel_size=7, padding=3, bias=True)
        self.bn7 = torch.nn.InstanceNorm2d(in_channels)

        self.leaky_relu = torch.nn.LeakyReLU(negative_slope=0.2)  # 设置负斜率为 0.01

        self.score_conv = torch.nn.Conv2d(in_channels=in_channels * 3, out_channels=in_channels, kernel_size=1, bias=True)
        self.bnscore = torch.nn.InstanceNorm2d(in_channels)


    def forward(self, input):
        epsilon = 1e-6
        edge_x = self.sobel_h(input)
        edge_y = self.sobel_v(input)
        angles = torch.atan2(edge_y, edge_x + epsilon)
        sobel = torch.cat((edge_x, edge_y, angles), dim=1)
        sobel = self.sobelbn(sobel)




        edge3_x = self.sobel3_x(input)
        edge3_y = self.sobel3_y(input)
        angles3 = torch.atan2(edge3_y, edge3_x + epsilon)
        sobel3 = torch.cat((edge3_x, edge3_y, angles3), dim=1)
        sobel3 = self.sobelbn3(sobel3)



        edge5_x = self.sobel5_x(input)
        edge5_y = self.sobel5_y(input)
        angles5 = torch.atan2(edge5_y, edge5_x + epsilon)
        sobel5 = torch.cat((edge5_x, edge5_y, angles5), dim=1)
        sobel5 = self.sobelbn5(sobel5)

        edge7_x = self.sobel7_x(input)
        edge7_y = self.sobel7_y(input)
        angles7 = torch.atan2(edge7_y, edge7_x + epsilon)
        sobel7 = torch.cat((edge7_x, edge7_y, angles7), dim=1)
        sobel7 = self.sobelbn7(sobel7)

        x = torch.cat((sobel, sobel3, sobel5, sobel7), dim=1)
        x = x / x.size(1)
        ca = self.CA(x)
        sa = self.SA(x)

        csa = ca + sa

        pa = self.PA(x, csa)

        x = x + pa




        # batch_size, channels, height, width = x.size()
        # 整合不同感受野的信息
        attention_33 = self.leaky_relu(self.bn3(self.conv33(x)))
        attention_55 = self.leaky_relu(self.bn5(self.conv55(x)))
        attention_77 = self.leaky_relu(self.bn7(self.conv77(x)))
        attention_core = self.bnscore(self.score_conv(torch.cat((attention_33, attention_55, attention_77), dim=1)))
        # output = torch.sigmoid(attention_core)
        return attention_core




from PepperPepper.environment import torch
from PepperPepper.models.TransUNet.Transformer import Transformer
from PepperPepper.models.TransUNet.DecoderCup import DecoderCup
from PepperPepper.models.TransUNet.SegmentationHead import SegmentationHead


class VisionTransformer(torch.nn.Module):
    def __init__(self, config, img_size=224, num_classes=None, zero_head=False, vis=False):
        super(VisionTransformer, self).__init__()
        if num_classes==None:
            num_classes = config.n_classes
        else:
            num_classes = num_classes
        self.num_classes = num_classes
        self.zero_head = zero_head
        self.classifier = config.classifier
        self.transformer = Transformer(config, img_size, vis)
        self.decoder = DecoderCup(config)
        self.segmentation_head = SegmentationHead(
            in_channels=config['decoder_channels'][-1],
            out_channels=self.num_classes,
            kernel_size=3,
        )
        self.config = config




    def forward(self, x):
        if x.size()[1] == 1:
            x = x.repeat(1,3,1,1)
        x, attn_weights, features = self.transformer(x)  # (B, n_patch, hidden)
        x = self.decoder(x, features)
        logits = self.segmentation_head(x)
        return logits


from PepperPepper.models.TransUNet.ResNetV2 import ResNetV2
from PepperPepper.environment import torch, collections, itertools

def _ntuple(n, name="parse"):
    def parse(x):
        if isinstance(x, collections.abc.Iterable):
            return tuple(x)
        return tuple(itertools.repeat(x, n))

    parse.__name__ = name
    return parse


_single = _ntuple(1, "_single")
_pair = _ntuple(2, "_pair")
_triple = _ntuple(3, "_triple")
_quadruple = _ntuple(4, "_quadruple")


class Embeddings(torch.nn.Module):
    """Construct the embeddings from patch, position embeddings.
    """
    def __init__(self, config, img_size, in_channels=3):
        super(Embeddings, self).__init__()
        self.hybrid = None
        self.config = config
        img_size = _pair(img_size)

        if config.patches.get("grid") is not None:   # ResNet
            grid_size = config.patches["grid"]
            patch_size = (img_size[0] // 16 // grid_size[0], img_size[1] // 16 // grid_size[1])
            patch_size_real = (patch_size[0] * 16, patch_size[1] * 16)
            n_patches = (img_size[0] // patch_size_real[0]) * (img_size[1] // patch_size_real[1])
            self.hybrid = True
        else:
            patch_size = _pair(config.patches["size"])
            n_patches = (img_size[0] // patch_size[0]) * (img_size[1] // patch_size[1])
            self.hybrid = False

        if self.hybrid:
            self.hybrid_model = ResNetV2(in_channels= in_channels,block_units=config.resnet.num_layers, width_factor=config.resnet.width_factor)
            in_channels = self.hybrid_model.width * 16
        self.patch_embeddings = torch.nn.Conv2d(in_channels=in_channels,
                                       out_channels=config.hidden_size,
                                       kernel_size=patch_size,
                                       stride=patch_size)
        self.position_embeddings = torch.nn.Parameter(torch.zeros(1, n_patches, config.hidden_size))

        self.dropout = torch.nn.Dropout(config.transformer["dropout_rate"])


    def forward(self, x):
        if self.hybrid:
            x, features = self.hybrid_model(x)
        else:
            features = None
        x = self.patch_embeddings(x)  # (B, hidden. n_patches^(1/2), n_patches^(1/2))
        x = x.flatten(2)
        x = x.transpose(-1, -2)  # (B, n_patches, hidden)

        embeddings = x + self.position_embeddings
        embeddings = self.dropout(embeddings)
        return embeddings, features


import numpy as np

from ..environment import torch, cv2, plt




'''
1.visualize_grid_cells(image_path, W, H, X, Y)
将图像划分的grid cell进行可视化
image_path：图像的路径
W:设置图像的宽
H：设置图像的高
X、Y：设置grid cell的数量为X行×Y列个
'''
def visualize_grid_cells(image_path, W, H, ROW, COLUMN, Title='Resized Image with Grid Cells'):
    # 读取图像
    image = cv2.imread(image_path)

    # 如果图像为空（例如文件不存在或损坏），则退出
    if image is None:
        print(f"Error: Could not read image at {image_path}")
        return

        # 调整图像大小
    resized_image = cv2.resize(image, (W, H))

    # 划分grid cell
    cell_width = W // COLUMN
    cell_height = H // ROW

    # 遍历每个grid cell，绘制边界
    for i in range(COLUMN):
        for j in range(ROW):
            # 计算grid cell的左上角坐标
            start_x = i * cell_width
            start_y = j * cell_height
            # 计算grid cell的右下角坐标
            end_x = min((i + 1) * cell_width, W)  # 防止超出图像宽度
            end_y = min((j + 1) * cell_height, H)  # 防止超出图像高度

            # 绘制grid cell边界（使用BGR颜色，因为在OpenCV中图像是BGR格式的）
            cv2.rectangle(resized_image, (start_x, start_y), (end_x, end_y), (0, 255, 0), 2)

            # 显示原图与grid cell（合并到一张图像）
    plt.imshow(cv2.cvtColor(resized_image, cv2.COLOR_BGR2RGB))
    plt.title(Title)
    plt.axis('off')
    plt.show()





'''
2.show_images(imgs, num_rows, num_cols, titles=None, scale=1.5, normalize = False)
展示图像
'''
def show_images(imgs, num_rows, num_cols, titles=None, scale=1.5, normalize = False):
    # 绘制图像列表
    figsize = (num_cols * scale, num_rows * scale)
    _, axes = plt.subplots(num_rows, num_cols, figsize=figsize)
    axes = axes.flatten()
    for i, (ax, img) in enumerate(zip(axes, imgs)):
        if torch.is_tensor(img):
            # 图像张量
            if img.dim() == 3:
                img = img.permute(1, 2, 0)  # 将通道维度移动到最后的位置
            if normalize:
                img = img.numpy().clip(0, 1)  # 归一化到0-1范围，并裁剪超出范围的值
            ax.imshow(img)

        else:
            # PIL图像
            ax.imshow(img)
        ax.axes.get_xaxis().set_visible(False)
        ax.axes.get_yaxis().set_visible(False)
        if titles:
            ax.set_title(titles[i])
    return axes





'''
3.box_corner_to_center(boxes)
    summary: it transform the box corners to center
'''
def box_corner_to_center(boxes):
    x1, y1, x2, y2 = boxes[:, 0], boxes[:, 1], boxes[:, 2], boxes[:, 3]
    cx = (x1 + x2) // 2
    cy = (y1 + y2) // 2
    w = x2 - x1
    h = y2 - y1
    boxes = torch.stack([cx, cy, w, h], axis=-1)
    return boxes





'''
4.bbox_to_rect(bbox, color)
    summary: it draw bounding box rectangle
'''
def bbox_to_rect(bbox, color):
    return plt.Rectangle(xy=(bbox[0], bbox[1]),width=bbox[2]-bbox[0],height=bbox[3]-bbox[1],fill=False,edgecolor=color,linewidth=2)


'''
5.show_boxes(axes, bboxes, labels=None, colors=None)
    summary: it show the box corners
    
'''
def show_boxes(axes, bboxes, labels, colors=None):
    """显示所有的边框"""
    def _make_list(obj, default_values=None):
        if obj is None:
            obj = default_values
        elif not isinstance(obj, (list, tuple)):
            obj = [obj]
        return obj

    labels = _make_list(labels)
    colors = _make_list(colors,['b', 'g', 'r', 'm', 'c' ])
    for i, bbox in enumerate(bboxes):
        color = colors[i % len(colors)]
        rect = bbox_to_rect(bbox.detach().numpy(), color)
        axes.add_patch(rect)
        if labels and len(labels)>i:
            text_color = 'k' if color== 'w' else 'w'
            axes.text(rect.xy[0],rect.xy[1],labels[i],va='center',ha='center',color=text_color,fontsize=9,bbox=dict(facecolor=color, lw=0))


# 5.visualize_bounding_boxes
def visualize_bounding_boxes(image, bounding_boxes, labels):
    """
    visualize_bounding_boxes(image, bounding_boxes, labels)
    Visualizes bounding boxes on an image.

    Parameters:
        image (numpy.ndarray): The input image.
        bounding_boxes (list): A list of bounding boxes in the format [xmin, ymin, xmax, ymax].
        labels (list): A list of labels corresponding to each bounding box.
    """
    # Create figure and axes
    fig, ax = plt.subplots()

    # Display the image
    ax.imshow(image)

    # Add bounding boxes
    for bbox, label in zip(bounding_boxes, labels):
        xmin, ymin, xmax, ymax = bbox
        # Convert relative coordinates to absolute coordinates
        xmin_abs = xmin * image.shape[1]
        ymin_abs = ymin * image.shape[0]
        xmax_abs = xmax * image.shape[1]
        ymax_abs = ymax * image.shape[0]
        width = xmax_abs - xmin_abs
        height = ymax_abs - ymin_abs
        # Create a Rectangle patch
        rect = plt.Rectangle((xmin_abs, ymin_abs), width, height, fill=False, edgecolor='red', linewidth=2)
        ax.add_patch(rect)
        # Add label
        ax.text(xmin_abs, ymin_abs, label, color='red', fontsize=12, verticalalignment='top')

    plt.show()



'''
6.box_iou_xyxy()
    summary: 计算预测框与真实框之间的IOU值，坐标格式为xyxy
    Args:
        boxes (list): A list of bounding boxes in the format [xmin, ymin, xmax, ymax].
'''
def box_iou_xyxy(boxe1, boxe2):
    # 获取box1的具体坐标
    x1min, y1min, x2max, y2max = boxe1[0], boxe1[1], boxe1[2], boxe1[3]

    # 计算boxe1的框面积
    S1 = np.maximum((x2max - x1min) * (y2max - y1min), 0)

    # 获取box2的具体坐标
    x2min, y2min, x2max, y2max = boxe2[0], boxe2[1], boxe2[2], boxe2[3]

    # 计算boxe2的框面积.
    S2 = np.maximum((x2max - x2min) * (y2max - y2min), 0)

    # 计算相交矩形框的坐标，即boxe1与boxe2的交集
    xmin = np.maximum(x1min, x2min)
    ymin = np.maximum(y1min, y2min)
    xmax = np.minimum(x2max, x2max)
    ymax = np.minimum(y2max, y2max)

    # 计算两个矩形框的交集
    intersection = np.maximum((xmax - xmin) * (ymax - ymin), 0)

    IoU = intersection / np.maximum((S1 + S2 - intersection),1.)

    return IoU



'''
7.box_iou_xywh(boxe1, boxe2)
'''
def box_iou_xywh(boxe1, boxe2):
    # 将xywh化为xyxy
    x1min, y1min, x1max, y1max = boxe1[0] - boxe1[2]/2.0 , boxe1[1] - boxe1[3]/2.0, boxe1[0] + boxe1[2]/2.0 , boxe1[1] + boxe1[3]/2.0
    x2min, y2min, x2max, y2max = boxe2[0] - boxe2[2]/2.0 , boxe2[1] - boxe2[3]/2.0, boxe2[0] + boxe2[2]/2.0 , boxe2[1] + boxe2[3]/2.0

    return box_iou_xyxy((x1min, y1min, x1max, y1max), (x2min, y2min, x2max, y2max))




'''
8.draw_anchor_boxes(image, center_x, center_y, widths, heights)
'''
def draw_anchor_boxes(image, center_x, center_y, widths, heights):
    """
    在图像中画出给定中心点和多种宽高的锚框。

    参数：
        image: 输入的图像，可以是 numpy 数组。
        center_x: 中心点的 x 坐标。
        center_y: 中心点的 y 坐标。
        widths: 锚框的宽度列表。
        heights: 锚框的高度列表。

    返回值：
        annotated_image: 画有锚框的图像，numpy 数组格式。
    """
    # 创建一个副本以保留原始图像
    annotated_image = image.copy()

    # 循环绘制每个锚框
    for width, height in zip(widths, heights):
        # 计算左上角和右下角坐标
        top_left = (int(center_x - width / 2), int(center_y - height / 2))
        bottom_right = (int(center_x + width / 2), int(center_y + height / 2))

        # 在图像上画出锚框
        annotated_image = cv2.rectangle(annotated_image, top_left, bottom_right, (0, 255, 0), 2)

    return annotated_image









'''
9.get_objectness_label(img, gt_boxes, gtlabels, iou_threshold=0.7, anchors = [])
    summary: 标注预测框的objectness
    img 是输入的图像数据，形状是[N, C, H, W]
    gt_boxes，真实框，维度是[N, 50, 4]，其中50是真实框数目的上限，当图片中真实框不足50个时，不足部分的坐标全为0
              真实框坐标格式是xywh，这里使用相对值
    gt_labels，真实框所属类别，维度是[N, 50]
    iou_threshold，当预测框与真实框的iou大于iou_threshold时不将其看作是负样本
    anchors，锚框可选的尺寸
    anchor_masks，通过与anchors一起确定本层级的特征图应该选用多大尺寸的锚框
    num_classes，类别数目
    downsample，特征图相对于输入网络的图片尺寸变化的比例
'''
def get_objectness_label(img, gt_boxes, gtlabels, iou_threshold=0.7, anchors = [116, 90, 156, 198, 373, 326], num_classes = 3, downsample = 32):
    img_shape = img.shape
    batch_size = img.shape[0]
    num_anchors = len(anchors) // 2
    input_h = img.shape[2]
    input_w = img.shape[3]

    # 将输入图片划分成num_rows x num_cols个候选区域，每个候选区域的边长是downsample
    # 计算共有多少列候选区域
    num_cols = input_w // downsample
    # 计算共有多少行候选区域
    num_rows = input_h // downsample

    label_objectness = np.zeros([batch_size, num_anchors, num_rows, num_cols])
    label_classification = np.zeros([batch_size, num_anchors, num_classes, num_rows, num_cols])
    label_location = np.zeros([batch_size, num_anchors, 4, num_rows, num_cols])

    # 这里有个scale_location，用来调节大目标和小目标的权重的
    # 大目标和小目标对于相同的偏移，大目标更加敏感，故增强这个权重系数
    scale_location = np.zeros([batch_size, num_anchors, num_rows, num_cols])

    # batch_size进行循环，处理每一张图片
    for n in range(batch_size):
        # 对图片上的真实框进行遍历，依次找出与真实框形状最匹配的锚框
        for n_gt in range(len(gt_boxes[n])):
            gt_box = gt_boxes[n][n_gt]
            # 真实框所属类别
            gt_label = gtlabels[n][n_gt]
            # 真实框的中心坐标和长宽（YOLO中采用xywh）
            gt_center_x = gt_box[0]
            gt_center_y = gt_box[1]
            gt_width = gt_box[2]
            gt_height = gt_box[3]
            if (gt_width < 1e-3) or (gt_height < 1e-3):
                continue

            # 计算这个真实框的中心点落在哪个格子里,xy是真实边界框的中心点在图像中的归一化坐标（即范围在 [0, 1] 内）
            i = int(gt_center_y * num_rows)
            j = int(gt_center_x * num_cols)

            ious = []

            for ka in range(num_anchors):
                # x，y为0，假设锚框和真实框的x，y相同
                gt_bbox = [0., 0., float(gt_width), float(gt_height)]
                # 生成输入（i，j）这个格子里的锚框
                anchor_w = anchors[ka * 2]
                anchor_h = anchors[ka * 2 + 1]
                anchor_bbox = [0., 0., anchor_w / float(input_w), anchor_h / float(input_h)]
                # 计算iou
                iou = box_iou_xywh(gt_bbox, anchor_bbox)
                ious.append(iou)
            ious = np.array(ious)
            inds = np.argsort(ious)
            k = inds[-1]
            label_objectness[n, k, i, j] = 1
            c = gt_label
            # 使用one-hot方法标注
            label_classification[n, k, c, i, j] = 1

            # for those prediction bbox with objectness =1, set label of location
            # 反算出真实框的tx*，ty*，tw*，th*
            dx_label = gt_center_x * num_cols - j
            dy_label = gt_center_y * num_rows - i
            dw_label = np.log(gt_width * input_w / anchors[k * 2])
            dh_label = np.log(gt_height * input_h / anchors[k * 2 + 1])
            label_location[n, k, 0, i, j] = dx_label
            label_location[n, k, 1, i, j] = dy_label
            label_location[n, k, 2, i, j] = dw_label
            label_location[n, k, 3, i, j] = dh_label

            # scale_location用来调节不同尺寸的锚框对损失函数的贡献，作为加权系数和位置损失函数相乘
            # 根据这个计算方法，可以得出真实框越大，权重越小
            scale_location[n, k, i, j] = 2.0 - gt_width * gt_height

    # 目前根据每张图片上所有出现过的gt_box都标注出了objectness为正的预测框，剩下的预测框的objectness为0
    # 对于objectness为1的预测框，标出它们所包含的物体类别以及位置回归的目标
    return label_objectness.astype('float32'), label_location.astype('float32'), label_classification.astype('float32'), scale_location.astype('float32')






'''
10.nms(bboxes, scores, score_thresh, nms_thresh, pre_nms_topk, i=0, c=0)
    NMS非极大值抑制
    summary: Non-Maximum Suppression
'''
def nms(bboxes, scores, score_thresh, nms_thresh, pre_nms_topk, i=0, c=0):
    inds = np.argsort(scores)
    inds = inds[::-1]
    keep_inds = []
    while(len(inds) > 0):
        cur_ind = inds[0]
        cur_score = scores[cur_ind]
        # 如果该框的分数小于score.thresh，则丢弃该框
        if cur_score < score_thresh:
            break

        keep = True
        for ind in keep_inds:
            current_box = bboxes[cur_ind]
            remain_box = bboxes[ind]
            iou = box_iou_xywh(current_box, remain_box)
            if iou > nms_thresh:
                keep = False
                break
        # 调试代码，用于观察信息使用
        if i == 0 and c == 4 and cur_ind == 951:
            print('suppressed, ',keep,i,c,cur_ind,ind,iou)
        if keep:
            keep_inds.append(cur_ind)
        # 去除当前处理的边界框索引，即将 inds 数组中的第一个元素移除，继续处理下一个边界框。
        inds = inds[1:]
    return np.array(keep_inds)






'''
11.multiclass_nms(bboxes, scores, score_thresh = 0.01, nms_thresh = 0.45, pre_nms_topk = 1000, pos_nms_topk = 100 )
    summary: multiclass-Non-Maximum Suppression
'''
def multiclass_nms(bboxes, scores, score_thresh = 0.01, nms_thresh = 0.45, pre_nms_topk = 1000, pos_nms_topk = 100 ):
    # 获取批量大小和类别数量
    batch_size = bboxes.shape[0]
    class_num = scores.shape[1]
    rets = []
    # 遍历每个样本
    for i in range(batch_size):
        # 获取当前样本的边界框和分数
        bboxes_i = bboxes[i]
        scores_i = scores[i]
        ret = []
        # 遍历每个类别
        for c in range(class_num):
            # 获取当前类别的分数
            scores_i_c = scores_i[c]
            # 使用 NMS 函数进行非最大抑制
            keep_inds = nms(bboxes_i, scores_i_c, score_thresh, nms_thresh, pre_nms_topk, i, c)

            # 如果保留的边界框数量小于1，则跳过当前类别
            if len(keep_inds) < 1:
                continue
            # 根据保留的索引获取对应的边界框和分数
            keep_bboxes = bboxes_i[keep_inds]
            keep_scores = scores_i_c[keep_inds]
            # 构造保留结果
            keep_results = np.zeros([keep_scores.shape[0], 6])
            keep_results[:, 0] = c # 类别索引
            keep_results[:, 1] = keep_scores[:] # 分数
            keep_results[:, 2:6] = keep_bboxes[:, :]    # 边界框坐标
            ret.append(keep_results)
        # 如果当前样本没有保留的结果，则跳过
        if len(ret) < 1:
            rets.append(ret)
            continue
        # 将每个类别的结果拼接在一起
        ret_i = np.concatenate(ret, axis=0)
        # 如果保留的结果数量超过了指定的阈值，则根据分数进行排序并截取指定数量的结果
        scores_i = ret_i[:, 1]
        if len(scores_i) > pos_nms_topk:
            inds = np.argsort(scores_i)[::-1]
            inds = inds[:pos_nms_topk]
            ret_i = ret_i[inds]

        rets.append(ret_i)

    return rets



'''
12.get_yolo_box_xxyy(pred, anchors, num_classes, downsample)
    summary: get_yolo_box_xxyy将网络特征图输出的[tx,ty,tw,th]转化为预测框的坐标[x1,y1,x2,y2]
    
'''
def get_yolo_box_xxyy(pred, anchors, num_classes, downsample):
    """
    Args:
    :param pred: 网络输出特征图,[N,C,H,W],其中C = NUM_ANCHORS * (5 + NUM_CLASSES),类型为numpy.ndarray
    :param anchors: list,表示锚框的大小
    :param num_classes: 类别数
    :param downsample: 相比于原图下采样多少倍
    :return: pred_box: 预测框的坐标
    """
    batch_size = pred.shape[0]
    num_rows = pred.shape[-2]
    num_cols = pred.shape[-1]

    input_h = num_rows * downsample
    input_w = num_cols * downsample

    num_anchors = len(anchors)//2

    # 对pred进行改造
    pred = pred.reshape([-1, num_anchors, 5+num_classes, num_rows, num_cols])
    pred_location = pred[:, :, 0:4, :, :]
    pred_location = torch.permute(pred_location, (0, 3, 4, 1, 2))
    anchors_this = []
    for ind in range(num_anchors):
        anchors_this.append([anchors[ind*2], anchors[ind*2 + 1]])
    anchors_this = torch.tensor(anchors_this).float()

    # 最终输出数据保存在pred_box中，其形状是[N, H, W, NUM_ANCHORS, 4],
    # 其中最后一个维度4代表位置的4个坐标
    pred_box = torch.zeros(pred_location.shape)
    for n in range(batch_size):
        for i in range(num_rows):
            for j in range(num_cols):
                for k in range(num_anchors):
                    pred_box[n,i,j,k,0] = j
                    pred_box[n,i,j,k,1] = i
                    pred_box[n,i,j,k,2] = anchors_this[k][0]
                    pred_box[n,i,j,k,3] = anchors_this[k][1]
    pred_box = pred_box.to(pred_location.device)


    # 使用相对坐标，pred_box的输出元素数值在0~1.0之间
    pred_box[:, :, :, :, 0] = (torch.nn.functional.sigmoid(pred_location[:, :, :, :, 0]) + pred_box[:, :, :, :, 0]) / num_cols
    pred_box[:, :, :, :, 1] = (torch.nn.functional.sigmoid(pred_location[:, :, :, :, 1]) + pred_box[:, :, :, :, 1]) / num_rows
    pred_box[:, :, :, :, 2] = torch.exp(pred_location[:, :, :, :, 2]) * pred_box[:, :, :, :, 2] / input_w
    pred_box[:, :, :, :, 3] = torch.exp(pred_location[:, :, :, :, 3]) * pred_box[:, :, :, :, 3] / input_h

    # 将坐标从xywh转化成xyxy
    pred_box[:, :, :, :, 0] = pred_box[:, :, :, :, 0] - pred_box[:, :, :, :, 2] / 2.
    pred_box[:, :, :, :, 1] = pred_box[:, :, :, :, 1] - pred_box[:, :, :, :, 3] / 2.
    pred_box[:, :, :, :, 2] = pred_box[:, :, :, :, 0] + pred_box[:, :, :, :, 2]
    pred_box[:, :, :, :, 3] = pred_box[:, :, :, :, 1] + pred_box[:, :, :, :, 3]

    pred_box = torch.clip(pred_box, 0., 1.0)
    return pred_box










'''
13.image_downsampling(image, ratio)
    summary: 对图像进行下采样。
    Parameters:
        image: shape：(w,h,c),type：numpy.ndarray
        ratio: 下采样比率，ratio>1且最好为整数
        manner: 采样方式,默认为mean
    Return:
        image_sampling(numpy.ndarray):shape:(w,h,c),type:numpy.ndarray
        
'''
def image_downsampling(image, ratio, manner='mean'):
    image_sampling = np.zeros((int(image.shape[0]/ratio), int(image.shape[1]/ratio), image.shape[2]), dtype='int32')
    fun = np.mean
    if manner == 'mean':
        fun = np.mean
    elif manner == 'max':
        fun = np.max
    for i in range(image_sampling.shape[0]):
        for j in range(image_sampling.shape[1]):
            for k in range(image_sampling.shape[2]):
                delta = image[i * ratio:(i + 1) * ratio, j * ratio:(j + 1) * ratio, k]
                image_sampling[i, j, k] = fun(delta)

    return image_sampling

'''
14.image_quantization(image, ratio)
    summary: 对图像进行量化
    Parameters:
        image: shape：(w,h,c),type：numpy.ndarray
        ratio: 设置量化比率
    Returns:
        image: shape：(w,h,c),type：numpy.ndarray,量化后的图像数据
'''
def image_quantization(image, ratio):
    image_quantization = image.copy()
    for i in range(image_quantization.shape[0]):
        for j in range(image_quantization.shape[1]):
            for k in range(image_quantization.shape[2]):
                image_quantization[i][j][k] = int(image_quantization[i][j][k] / ratio) * ratio

    return image_quantization





'''
15.image_Brightness(image, value)
    summary: 对图像进行亮度调整
    Parameters:
        image为shape：(w,h,c),type：numpy.ndarray
        value：为int类型，可正可负，正为增强图像的亮度，负为降低图像的亮度
    Return：
        image_Brightness：(w,h,c),type：numpy.ndarray
'''
def image_Brightness(image, value):
    image_Brightness = image.copy()
    for i in range(image_Brightness.shape[0]):
        for j in range(image_Brightness.shape[1]):
            for k in range(image_Brightness.shape[2]):
                image_Brightness[i][j][k] = min(image_Brightness[i][j][k], image_Brightness[i][j][k] + value)

    return image_Brightness





'''
16.image_contrast(image, ratio)
    summary: 对图像的对比度
    Parameters:
        image,shape：(w,h,c),type：numpy.ndarray
        ratio,所有元素乘以这个系数增大最大灰度值与最小灰度值的差异即落差值
    Return:
        image_contrast,(w,h,c),type：numpy.ndarray
'''
def image_contrast(image, ratio):
    image_contrast = image.copy()
    for i in range(image_contrast.shape[0]):
        for j in range(image_contrast.shape[1]):
            for k in range(image_contrast.shape[2]):
                if image_contrast[i,j,k]*ratio > 255:
                    image_contrast[i,j,k] = 255
                elif image_contrast[i,j,k]*ratio < 0:
                    image_contrast[i,j,k] = 0
                else:
                    image_contrast[i,j,k] = image_contrast[i,j,k] * ratio
    return image_contrast



'''
17.image_cv2plt(image)
    将图像bgr空间转rgb空间
'''
def image_cv2plt(image):
    imgCp = image.copy()
    imgCp = cv2.cvtColor(imgCp, cv2.COLOR_BGR2RGB)
    return imgCp






from ..environment import re, os


'''
1.read_time_machine()
    将时间机器数据集加载到文本行列表中
    
    Parameters:
    - None

    Returns:
    - list: 文本行列表
'''
def read_time_machine():
    from .dataset_utils import download
    """将时间机器数据集加载到文本行的列表中"""
    with open(download('time_machine'), 'r') as f:
        lines = f.readlines()
    return [re.sub('[^A-Za-z]+', ' ', line).strip().lower() for line in lines]




'''
2.load_corpus_time_machine(max_tokens=-1)
    返回时光机器数据集的词元索引列表和词表
    
    Parameters:
    - max_tokens: int，corpus取多少数
    
    Returns:
    - corpus(list) : 词元索引列表
    - vocab(Vocab) : 时光机器语料库的词表
'''
def load_corpus_time_machine(max_tokens=-1):
    from .dataset_utils import tokenize, Vocab
    lines = read_time_machine()
    tokens = tokenize(lines,'char')
    vocab = Vocab(tokens)

    # 将时光机器数据集文本展平到一个列表中
    corpus = [vocab[token] for line in tokens for token in line]
    if max_tokens > 0:
        corpus = corpus[:max_tokens]
    return corpus, vocab


'''
3.SeqDataLoader
    加载序列数据的迭代器。这个类提供了两种加载序列数据的方式：随机抽样（seq_data_iter_random）和顺序抽取（seq_data_iter_sequential）。

    Parameters:
    - batch_size (int): 每个小批量中的样本数量。  
    - num_steps (int): 每个子序列的长度。  
    - use_random_iter (bool): 是否使用随机迭代。  
    - max_tokens (int): 用于加载语料库时的最大词元数。  

    Attributes:
    - data_iter_fn (callable): 用于生成数据迭代器的函数（随机或顺序）。
    - corpus (list): 加载的语料库，包含序列数据。
    - vocab (dict): 词汇表，将词元映射到索引。
    - batch_size (int): 每个小批量中的样本数量。
    - num_steps (int): 每个子序列的长度。
'''


class SeqDataLoader_time_machine:
    """加载序列数据的迭代器类"""

    def __init__(self, batch_size, num_steps, use_random_iter, max_tokens):
        """
        初始化序列数据加载器

        Args:
            batch_size (int): 每个小批量中的样本数量。
            num_steps (int): 每个子序列的长度。
            use_random_iter (bool): 是否使用随机迭代。
            max_tokens (int): 用于加载语料库时的最大词元数。

        Attributes:
            data_iter_fn (callable): 用于生成数据迭代器的函数（随机或顺序）。
            corpus (list): 加载的语料库，包含序列数据。
            vocab (dict): 词汇表，将词元映射到索引。
            batch_size (int): 每个小批量中的样本数量。
            num_steps (int): 每个子序列的长度。
        """
        from .dataset_utils import seq_data_iter_random,seq_data_iter_sequential

        if use_random_iter:
            # 如果使用随机迭代，则将 data_iter_fn 设置为 seq_data_iter_random 函数
            self.data_iter_fn = seq_data_iter_random
        else:
            # 否则，将 data_iter_fn 设置为 seq_data_iter_sequential 函数
            self.data_iter_fn = seq_data_iter_sequential

            # 加载语料库和词汇表，并限制最大词元数为 max_tokens
        self.corpus, self.vocab = load_corpus_time_machine(max_tokens)

        # 设置 batch_size 和 num_steps
        self.batch_size, self.num_steps = batch_size, num_steps

    def __iter__(self):
        """
        实现迭代器协议，使得对象可迭代

        Returns:
            generator: 返回一个生成器，该生成器使用 data_iter_fn 函数生成小批量数据。
        """
        # 调用之前设置好的 data_iter_fn 函数来生成数据迭代器
        # 并返回该迭代器，用于迭代获取小批量数据
        return self.data_iter_fn(self.corpus, self.batch_size, self.num_steps)



'''
4.load_data_time_machine(batch_size, num_steps, use_random_iter=False, max_tokens=10000)
    返回时光机器数据集的迭代器和词表
    Parameters:
    - batch_size (int): 每个小批量中的样本数量。  
    - num_steps (int): 序列中每个样本的时间步数（或长度）。  
    - use_random_iter (bool, optional): 是否使用随机迭代来生成小批量。默认为False。  
    - max_tokens (int, optional): 加载语料库时使用的最大词元数。默认为10000。 

    Returns:  
    - data_iter (SeqDataLoader): 数据加载器的迭代器，用于生成小批量数据。  
    - vocab (dict): 词汇表，将词元映射到索引。  
'''
def load_data_time_machine(batch_size, num_steps, use_random_iter=False, max_tokens=10000):
    """返回时光机器数据集的迭代器和词表"""
    data_iter = SeqDataLoader_time_machine(
        batch_size, num_steps, use_random_iter, max_tokens)
    return data_iter, data_iter.vocab



'''
5.read_data_nmt()
    载入“英语－法语”数据集  
    该函数从指定的数据源下载并解压“英语－法语”数据集，然后读取法语部分（假设是fra.txt文件）并返回其内容。  
  
    Returns:  
        str: 法语文本数据，作为字符串返回。
'''
def read_data_nmt():
    """载入“英语－法语”数据集"""
    from .dataset_utils import download_extract
    data_dir = download_extract('fra-eng')
    with open(os.path.join(data_dir, 'fra.txt'), 'r',
             encoding='utf-8') as f:
        return f.read()






'''
6.preprocess_nmt(text)
    预处理“英语－法语”数据集，大写转小写，字母与符号之间加空格，多个空格转为一个空格符。
    
    Parameters:
    - text：读取的文本list。
    
    Returns:
    - list: 处理后的文本序列。
'''
def preprocess_nmt(text):
    """预处理“英语－法语”数据集"""
    def no_space(char, prev_char):
        return char in set(',.!?') and prev_char != ' '

    # 使用空格替换不间断空格
    # 使用小写字母替换大写字母
    text = text.replace('\u202f', ' ').replace('\xa0', ' ').lower()
    # 在单词和标点符号之间插入空格
    out = [' ' + char if i > 0 and no_space(char, text[i - 1]) else char
           for i, char in enumerate(text)]
    return ''.join(out)



'''
7.tokenize_nmt(text, num_examples=None)
    该函数用于对“英语－法语”翻译任务的数据集进行词元化。  
  
    Args:  
        text (str): 包含“英语－法语”翻译对的字符串，其中每行以换行符（'\n'）分隔，  
                     每行内部使用制表符（'\t'）分隔英语和法语文本。  
        num_examples (int, optional): 需要处理的翻译对数量。如果设置了此参数，  
                                      函数将只处理前`num_examples`个翻译对。默认为None，  
                                      表示处理所有翻译对。  
  
    Returns:  
        tuple: 包含两个列表的元组，分别表示源语言（英语）和目标语言（法语）的词元化结果。  
        - source (list): 源语言（英语）文本的词元化列表，其中每个元素都是一个字符串列表，每个字符串表示一个单词。  
        - target (list): 目标语言（法语）文本的词元化列表，与源语言列表的结构相同。
'''
def tokenize_nmt(text, num_examples=None):
    """词元化“英语－法语”数据数据集"""
    source, target = [], []
    for i, line in enumerate(text.split('\n')):
        if num_examples and i > num_examples:
            break
        parts = line.split('\t')
        if len(parts) == 2:
            source.append(parts[0].split(' '))
            target.append(parts[1].split(' '))
    return source, target





'''
8.load_data_nmt(batch_size, num_steps, num_examples=600)
    加载并预处理机器英-法翻译数据集，返回数据迭代器以及源语言和目标语言的词表。  
  
    Args:  
        batch_size (int): 每个小批量中的样本数。  
        num_steps (int): 序列的最大长度，用于截断或填充。  
        num_examples (int, optional): 用于加载的数据样本数。默认为600。  
  
    Returns:  
        Tuple[DataLoader, Vocab, Vocab]:  
        - data_iter (DataLoader): 加载数据的小批量迭代器。  
        - src_vocab (Vocab): 源语言词表。  
        - tgt_vocab (Vocab): 目标语言词表。  
'''
def load_data_nmt(batch_size, num_steps, num_examples=600):
    from .dataset_utils import Vocab, build_array_nmt
    from .custom_dataset import load_arrays
    """返回翻译数据集的迭代器和词表"""
    text = preprocess_nmt(read_data_nmt())
    source, target = tokenize_nmt(text, num_examples)
    src_vocab = Vocab(source, min_freq=2,
                          reserved_tokens=['<pad>', '<bos>', '<eos>'])
    tgt_vocab = Vocab(target, min_freq=2,
                          reserved_tokens=['<pad>', '<bos>', '<eos>'])
    src_array, src_valid_len = build_array_nmt(source, src_vocab, num_steps)
    tgt_array, tgt_valid_len = build_array_nmt(target, tgt_vocab, num_steps)
    data_arrays = (src_array, src_valid_len, tgt_array, tgt_valid_len)
    data_iter = load_arrays(data_arrays, batch_size)
    return data_iter, src_vocab, tgt_vocab


from PepperPepper.environment import torch, np




class Conv2dReLU(torch.nn.Sequential):
    def __init__(
            self,
            in_channels,
            out_channels,
            kernel_size,
            padding=0,
            stride=1,
            use_batchnorm=True,
    ):
        conv = torch.nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            bias=not (use_batchnorm),
        )
        relu = torch.nn.ReLU(inplace=True)

        bn = torch.nn.BatchNorm2d(out_channels)

        super(Conv2dReLU, self).__init__(conv, bn, relu)




class DecoderBlock(torch.nn.Module):
    def __init__(
            self,
            in_channels,
            out_channels,
            skip_channels=0,
            use_batchnorm=True,
    ):
        super().__init__()
        self.conv1 = Conv2dReLU(
            in_channels + skip_channels,
            out_channels,
            kernel_size=3,
            padding=1,
            use_batchnorm=use_batchnorm,
        )
        self.conv2 = Conv2dReLU(
            out_channels,
            out_channels,
            kernel_size=3,
            padding=1,
            use_batchnorm=use_batchnorm,
        )
        self.up = torch.nn.UpsamplingBilinear2d(scale_factor=2)

    def forward(self, x, skip=None):
        x = self.up(x)
        if skip is not None:
            x = torch.cat([x, skip], dim=1)
        x = self.conv1(x)
        x = self.conv2(x)
        return x



class DecoderCup(torch.nn.Module):
    def __init__(self, config):
        super().__init__()
        self.config = config
        head_channels = 512
        self.conv_more = Conv2dReLU(
            config.hidden_size,
            head_channels,
            kernel_size=3,
            padding=1,
            use_batchnorm=True,
        )
        decoder_channels = config.decoder_channels
        in_channels = [head_channels] + list(decoder_channels[:-1])
        out_channels = decoder_channels

        if self.config.n_skip != 0:
            skip_channels = self.config.skip_channels
            for i in range(4-self.config.n_skip):  # re-select the skip channels according to n_skip
                skip_channels[3-i]=0

        else:
            skip_channels=[0,0,0,0]

        blocks = [
            DecoderBlock(in_ch, out_ch, sk_ch) for in_ch, out_ch, sk_ch in zip(in_channels, out_channels, skip_channels)
        ]
        self.blocks = torch.nn.ModuleList(blocks)

    def forward(self, hidden_states, features=None):
        B, n_patch, hidden = hidden_states.size()  # reshape from (B, n_patch, hidden) to (B, h, w, hidden)
        h, w = int(np.sqrt(n_patch)), int(np.sqrt(n_patch))
        x = hidden_states.permute(0, 2, 1)
        x = x.contiguous().view(B, hidden, h, w)
        x = self.conv_more(x)
        for i, decoder_block in enumerate(self.blocks):
            if features is not None:
                skip = features[i] if (i < self.config.n_skip) else None
            else:
                skip = None
            x = decoder_block(x, skip=skip)
        return x


from ..environment import torch, math


'''
1.train_epoch_ch8(net, train_iter, loss, updater, device, use_random_iter)
    parameter:
    - net: network to train
    - train_iter: iterator for training the network
    - loss: loss function
    - updater: updater is Optimizer
    - device: training which device is used
    - use_random_iter: whether to use random iterator for training
    
    Returns:
    - perplexity: perplexity of the network
    - speed: speed of training
'''

def train_epoch_ch8(net, train_iter, loss, updater, device, use_random_iter):
    '''动手学深度学习第八章，用于序列训练的训练函数'''
    from ..core.utils import Timer,Accumulator
    from ..optimizers.custom_optimizer import grad_clipping


    state, timer = None, Timer()
    metric = Accumulator(2)
    for X, Y in train_iter:
        if state is None or use_random_iter:
            # 在第一次迭代或使用随机抽样时初始化state
            state = net.begin_state(batch_size=X.shape[0], device=device)
        else:
            if isinstance(net, torch.nn.Module) and not isinstance(state, tuple):
                # state对于nn.GRU是个张量
                state.detach_()
            else:
                # state对于nn.LSTM或对于我们从零开始实现的模型是个张量
                for s in state:
                    s.detach_()
        y = Y.T.reshape(-1)
        X, y = X.to(device), y.to(device)
        y_hat, state = net(X, state)
        l = loss(y_hat, y.long()).mean()
        if isinstance(updater, torch.optim.Optimizer):
            updater.zero_grad()
            l.backward()
            grad_clipping(net, 1)
            updater.step()
        else:
            l.backward()
            grad_clipping(net, 1)
            # 因为已经调用了mean函数
            updater(batch_size=1)
        metric.add(l * y.numel(), y.numel())
    return math.exp(metric[0] / metric[1]), metric[1] / timer.stop()



'''
2.predict_ch8(prefix, num_preds, net, vocab, device)
在给定前缀后面生成新字符。

    Parameters:
        prefix (str): 用于生成新字符的前缀。
        num_preds (int): 预测的字符数。
        net (torch.nn.Module): 经过训练的神经网络模型。
        vocab (Vocab): 词汇表，将字符映射到索引。
        device (torch.device): 设备，用于计算。

    Returns:
        str: 生成的新字符序列。
'''
def predict_ch8(prefix, num_preds, net, vocab, device):  #@save
    """在prefix后面生成新字符"""
    state = net.begin_state(batch_size=1, device=device)
    outputs = [vocab[prefix[0]]]
    get_input = lambda: torch.tensor([outputs[-1]], device=device).reshape((1, 1))
    for y in prefix[1:]:  # 预热期
        _, state = net(get_input(), state)
        outputs.append(vocab[y])
    for _ in range(num_preds):  # 预测num_preds步
        y, state = net(get_input(), state)
        outputs.append(int(y.argmax(dim=1).reshape(1)))
    return ''.join([vocab.idx_to_token[i] for i in outputs])



'''
3.train_ch8(net, train_iter, vocab, lr, num_epochs, device,use_random_iter=False):
    训练模型（定义见第8章）
    
    参数:
        net (torch.nn.Module): 要训练的神经网络模型。
        train_iter (iterator): 训练数据集的迭代器。
        vocab (dict): 词汇表，用于将字符转换为索引。
        lr (float): 学习率，用于控制参数更新的步长。
        num_epochs (int): 训练的轮数。
        device (str): 训练设备，如'cpu'或'cuda:0'。
        use_random_iter (bool, optional): 是否使用随机数据迭代器，默认为False。
    
    返回值:
        无，直接打印训练过程中的困惑度和速度信息，并输出预测结果。
'''
def train_ch8(net, train_iter, vocab, lr, num_epochs, device,
              use_random_iter=False):

    from ..core.utils import Animator
    from ..optimizers import sgd
    """训练模型（定义见第8章）"""
    loss = torch.nn.CrossEntropyLoss()
    animator = Animator(xlabel='epoch', ylabel='perplexity',
                            legend=['train'], xlim=[10, num_epochs])
    # 初始化
    if isinstance(net, torch.nn.Module):
        updater = torch.optim.SGD(net.parameters(), lr)
    else:
        updater = lambda batch_size: sgd(net.params, lr, batch_size)
    predict = lambda prefix: predict_ch8(prefix, 50, net, vocab, device)
    # 训练和预测
    for epoch in range(num_epochs):
        ppl, speed = train_epoch_ch8(
            net, train_iter, loss, updater, device, use_random_iter)
        if (epoch + 1) % 10 == 0:
            print(predict('time traveller'))
            animator.add(epoch + 1, [ppl])
    print(f'困惑度 {ppl:.1f}, {speed:.1f} 词元/秒 {str(device)}')
    print(predict('time traveller'))
    print(predict('traveller'))










'''
4.train_seq2seq(net, data_iter, lr, num_epochs, tgt_vocab, device)
    序列模型训练函数
    Parameters:
        net： Seq模型
        data_iter： 数据集
        lr： 学习率
        num_epochs： 训练轮数
        tgt_vocab： 词表
        device： 训练设备
'''
def train_seq2seq(net, data_iter, lr, num_epochs, tgt_vocab, device):
    from ..core.utils import Animator, Timer, Accumulator
    from ..optimizers.custom_optimizer import grad_clipping
    from ..losses.custom_loss import MaskedSoftmaxCELoss
    """训练序列到序列模型"""
    def xavier_init_weights(m):
        if type(m) == torch.nn.Linear:
            torch.nn.init.xavier_uniform_(m.weight)
        if type(m) == torch.nn.GRU:
            for param in m._flat_weights_names:
                if "weight" in param:
                    torch.nn.init.xavier_uniform_(m._parameters[param])

    net.apply(xavier_init_weights)
    net.to(device)
    optimizer = torch.optim.Adam(net.parameters(), lr=lr)
    loss = MaskedSoftmaxCELoss()
    net.train()
    animator = Animator(xlabel='epoch', ylabel='loss',
                     xlim=[10, num_epochs])
    for epoch in range(num_epochs):
        timer = Timer()
        metric = Accumulator(2)  # 训练损失总和，词元数量
        for batch in data_iter:
            optimizer.zero_grad()
            X, X_valid_len, Y, Y_valid_len = [x.to(device) for x in batch]
            bos = torch.tensor([tgt_vocab['<bos>']] * Y.shape[0],
                          device=device).reshape(-1, 1)
            dec_input = torch.cat([bos, Y[:, :-1]], 1)  # 强制教学
            Y_hat, _ = net(X, dec_input, X_valid_len)
            l = loss(Y_hat, Y, Y_valid_len)
            l.sum().backward()      # 损失函数的标量进行“反向传播”
            grad_clipping(net, 1)
            num_tokens = Y_valid_len.sum()
            optimizer.step()
            with torch.no_grad():
                metric.add(l.sum(), num_tokens)
        if (epoch + 1) % 10 == 0:
            animator.add(epoch + 1, (metric[0] / metric[1],))
    print(f'loss {metric[0] / metric[1]:.3f}, {metric[1] / timer.stop():.1f} '
        f'tokens/sec on {str(device)}')





'''
5.predict_seq2seq(net, src_sentence, src_vocab, tgt_vocab, num_steps,device, save_attention_weights=False)
    Parameter:
        net: 预测模型
        src_sentence： 原句子
        src_vocab： 原词汇表
        tgt_vocab： 目标词汇表
        num_steps： 时间步
        device： 设备
        save_attention_weights： 保存注意力权重
'''
#@save
def predict_seq2seq(net, src_sentence, src_vocab, tgt_vocab, num_steps,
                    device, save_attention_weights=False):
    from ..datasets.dataset_utils import truncate_pad
    """序列到序列模型的预测"""
    # 在预测时将net设置为评估模式
    net.eval()
    src_tokens = src_vocab[src_sentence.lower().split(' ')] + [
        src_vocab['<eos>']]
    enc_valid_len = torch.tensor([len(src_tokens)], device=device)
    src_tokens = truncate_pad(src_tokens, num_steps, src_vocab['<pad>'])
    # 添加批量轴
    enc_X = torch.unsqueeze(
        torch.tensor(src_tokens, dtype=torch.long, device=device), dim=0)
    enc_outputs = net.encoder(enc_X, enc_valid_len)
    dec_state = net.decoder.init_state(enc_outputs, enc_valid_len)
    # 添加批量轴
    dec_X = torch.unsqueeze(torch.tensor(
        [tgt_vocab['<bos>']], dtype=torch.long, device=device), dim=0)
    output_seq, attention_weight_seq = [], []
    for _ in range(num_steps):
        Y, dec_state = net.decoder(dec_X, dec_state)
        # 我们使用具有预测最高可能性的词元，作为解码器在下一时间步的输入
        dec_X = Y.argmax(dim=2)
        pred = dec_X.squeeze(dim=0).type(torch.int32).item()
        # 保存注意力权重（稍后讨论）
        if save_attention_weights:
            attention_weight_seq.append(net.decoder.attention_weights)
        # 一旦序列结束词元被预测，输出序列的生成就完成了
        if pred == tgt_vocab['<eos>']:
            break
        output_seq.append(pred)
    return ' '.join(tgt_vocab.to_tokens(output_seq)), attention_weight_seq


import math

from PepperPepper.environment import torch, nn, F, rearrange

class ResidualBlock(nn.Module):
    def __init__(self, in_channels, out_channels, stride = 1):
        super(ResidualBlock, self).__init__()
        self.body = nn.Sequential(
            nn.Conv2d(in_channels, out_channels,  kernel_size=3, stride=stride, padding=1),
            nn.BatchNorm2d(out_channels),
            nn.LeakyReLU(0.2, inplace=True),

            nn.Conv2d(out_channels, out_channels, kernel_size=3, stride=stride, padding=1),
            nn.BatchNorm2d(out_channels),
        )
        if stride != 1 or out_channels != in_channels:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=stride),
                nn.BatchNorm2d(out_channels),
            )
        else:
            self.shortcut = None

        self.relu = nn.LeakyReLU(inplace=True)

    def forward(self, x):
        residual = x
        x = self.body(x)

        if self.shortcut is not None:
            residual = self.shortcut(residual)
        out = self.relu(x+residual)
        return out



class _FCNHead(torch.nn.Module):
    def __init__(self, in_channels, channels, norm_layer=torch.nn.BatchNorm2d):
        super(_FCNHead, self).__init__()
        self.block = torch.nn.Sequential()
        inter_channels = in_channels // 4
        self.block.append(torch.nn.Conv2d(in_channels=in_channels, out_channels=inter_channels, kernel_size=3, padding=1, bias=False))
        self.block.append(norm_layer(inter_channels))
        self.block.append(torch.nn.LeakyReLU(negative_slope=0.2))
        self.block.append(torch.nn.Dropout(0.1))
        self.block.append(torch.nn.Conv2d(in_channels=inter_channels, out_channels=channels, kernel_size=1))

    def forward(self, x):
        return self.block(x)



class Patch_embed(nn.Module):
    def __init__(self, in_channels, out_channels, patch_size, V='V1'):
        super().__init__()
        self.in_channels = in_channels
        self.patch_size = patch_size
        self.V = V
        self.out_channels = out_channels

        if self.V == 'V1' and self.is_power_of_two(patch_size):
            self.block = self._make_patch_embed_V1(in_channels, out_channels, patch_size)
        else:
            self.block = self._make_patch_embed_last(in_channels, out_channels, patch_size)
            # raise ValueError(f"Unsupported version: {self.V}")



    def forward(self, x):
        y = self.block(x)
        return y

    def is_power_of_two(self, n):
        if n <= 1:
            return False
        log2_n = math.log2(n)
        return log2_n.is_integer()

    def check_divisible_by_power_of_two(self, n, k):
        divisor = 2 ** k

        if n % divisor != 0:
            raise ValueError(f"Error: {n} is not divisible by 2^{k}. Please try again.")

        return


    def _make_patch_embed_V1(self, in_channels, out_channels, patch_size):
        stage_num = int(math.log2(patch_size))

        # self.check_divisible_by_power_of_two(out_channels, stage_num)
        dim = out_channels // stage_num



        block = []
        for d in range(stage_num):
            block.append(nn.Sequential(
                nn.Conv2d(in_channels * (d + 1) if d == 0 else dim * (d), dim * (d + 1) if d+1 != stage_num else out_channels, kernel_size=2, stride=2),
                Permute(0, 2, 3, 1),
                nn.LayerNorm(dim * (d + 1) if d+1 != stage_num else out_channels),
                Permute(0, 3, 1, 2),
                (nn.GELU() if d+1 != 0 else nn.Identity())
            ))

        return nn.Sequential(*block)


    def _make_patch_embed_last(self, in_channels, channels, patch_size):
        # block = []

        block = nn.Sequential(
            nn.Conv2d(in_channels, channels // 2, kernel_size=patch_size,stride=patch_size),
            Permute(0, 2, 3, 1),
            nn.LayerNorm(channels // 2),
            Permute(0, 3, 1, 2),
            nn.GELU(),
            nn.Conv2d(channels//2, channels, kernel_size=1, stride=1),
            Permute(0, 2, 3, 1),
            nn.LayerNorm(channels // 2),
            Permute(0, 3, 1, 2)
        )


        return block






class Permute(nn.Module):
    def __init__(self, *args):
        super().__init__()
        self.args = args

    def forward(self, x: torch.Tensor):
        return x.permute(*self.args)


class PatchExpand2D(nn.Module):
    def __init__(self, in_channels, out_channels, patch_size):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.patch_size = patch_size
        self.expand = nn.Linear(in_channels, patch_size*patch_size*self.out_channels, bias=False)
        self.norm = nn.LayerNorm(out_channels)

    def forward(self, x):#(b,h,w,c)->(b,h,w,2c)->(b,2h,2w,c/2)
        x=x.permute(0,2,3,1)
        B, H, W, C = x.shape
        x = self.expand(x)

        x = rearrange(x, 'b h w (p1 p2 c)-> b (h p1) (w p2) c', p1=self.patch_size, p2=self.patch_size, c=self.out_channels)
        x= self.norm(x).permute(0,3,1,2)

        return x




class FFT_PriorFilter(nn.Module):
    def __init__(self, in_channels , windows_size):
        super().__init__()
        self.dim = in_channels
        self.windows_size = windows_size
        self.cond = nn.Parameter(torch.randn(self.windows_size, self.windows_size // 2 + 1, self.dim, 2, dtype=(torch.float32)) * 0.02)


    def forward(self, x):
        B, C, H, W = x.size()
        assert H % self.windows_size == 0 and W % self.windows_size == 0, 'windows_size must be divisible by windows_size'
        x = rearrange(x, 'b c (h p1) (w p2) -> (b h w) p1 p2 c', p1=self.windows_size, p2=self.windows_size)

        X = torch.fft.rfft2(x, dim=(1, 2), norm='ortho')
        cond = torch.view_as_complex(self.cond)

        X = X * cond

        x = torch.fft.irfft2(X, s=(self.windows_size, self.windows_size), dim=(1, 2), norm='ortho')
        x = rearrange(x, '(b h w) p1 p2 c -> b c (h p1) (w p2)', h=H // self.windows_size, w=W // self.windows_size)

        return x








import numpy as np

from ..environment import torch, cv2, plt


'''
1.geometric_mean_filter
    summary: 几何均值滤波器,几何均值滤波器能够在一定程度上对图像进行降噪，且这种降噪使丢失的图像细节更少。其算法的基本原理是通过计算大小为 ｍ×ｎ的矩形子图像窗口的一组像素，每个复原的像素由子图像窗口中像素乘积的 １／ｍｎ次幂给出。
'''
def geometric_mean_filter(img, kernel_size=(3,3)):
    # kernel_size必须是奇数
    assert kernel_size[0] % 2 == 1 and kernel_size[1]%2 == 1, "Kernel size must be odd."

    img_h = img.shape[0]
    img_w = img.shape[1]
    m, n = kernel_size[0], kernel_size[1]
    order = 1 / (m * n)
    kernalMean = np.ones((m, n), np.float32)  # 生成盒式核

    hPad = int((m - 1) / 2)
    wPad = int((n - 1) / 2)
    imgPad = np.pad(img.copy(), ((hPad, m - hPad - 1), (wPad, n - wPad - 1)), mode="edge")
    imgGeoMean = img.copy()
    for i in range(hPad, img_h + hPad):
        for j in range(wPad, img_w + wPad):
            prod = np.prod(imgPad[i - hPad:i + hPad + 1, j - wPad:j + wPad + 1] * 1.0)
            imgGeoMean[i - hPad][j - wPad] = np.power(prod, order)

    return imgGeoMean






from ..environment import torch, plt
from ..environment import collections, math
from ..environment import random


'''
1.show_list_len_pair_hist(legend, xlabel, ylabel, xlist, ylist)
    绘制列表长度对的直方图，绘制文本序包含的词元数的直方图。
'''
def show_list_len_pair_hist(legend, xlabel, ylabel, xlist, ylist):
    from .utils import set_figsize
    """绘制列表长度对的直方图"""
    set_figsize()
    _, _, patches = plt.hist(
        [[len(l) for l in xlist], [len(l) for l in ylist]])
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    for patch in patches[1].patches:
        patch.set_hatch('/')
    plt.legend(legend)

'''
2.bleu(pred_seq, label_seq, k)
    Parameter:
        `pred_seq`：预测的序列（字符串形式）  
        `label_seq`：标签或参考序列（字符串形式）  
        `k`：n-gram的最大值
'''
def bleu(pred_seq, label_seq, k):
    """计算BLEU"""
    pred_tokens, label_tokens = pred_seq.split(' '), label_seq.split(' ')
    len_pred, len_label = len(pred_tokens), len(label_tokens)
    score = math.exp(min(0, 1 - len_label / len_pred))
    for n in range(1, k + 1):
        num_matches, label_subs = 0, collections.defaultdict(int)
        for i in range(len_label - n + 1):
            label_subs[' '.join(label_tokens[i: i + n])] += 1
        for i in range(len_pred - n + 1):
            if label_subs[' '.join(pred_tokens[i: i + n])] > 0:
                num_matches += 1
                label_subs[' '.join(pred_tokens[i: i + n])] -= 1
        score *= math.pow(num_matches / (len_pred - n + 1), math.pow(0.5, n))
    return score
































from PepperPepper.environment import threading, np, torch, F

__all__ = ['SegmentationMetricTPFNFP']

def get_miou_prec_recall_fscore(total_tp, total_fp, total_fn):
    miou = 1.0 * total_tp / (np.spacing(1) + total_tp + total_fp + total_fn)
    prec = 1.0 * total_tp / (np.spacing(1) + total_tp + total_fp)
    recall = 1.0 * total_tp / (np.spacing(1) + total_tp + total_fn)
    fscore = 2.0 * prec * recall / (np.spacing(1) + prec + recall)

    return miou, prec, recall, fscore

class SegmentationMetricTPFNFP(object):
    """Computes pixAcc and mIoU metric scroes
    """

    def __init__(self, nclass):
        self.nclass = nclass
        self.lock = threading.Lock()
        self.reset()

    def update(self, labels, preds):
        def evaluate_worker(self, label, pred):
            tp, fp, fn = batch_tp_fp_fn(pred, label, self.nclass)
            with self.lock:
                self.total_tp += tp
                self.total_fp += fp
                self.total_fn += fn
            return

        if isinstance(preds, torch.Tensor):
            preds = (preds.detach().numpy() > 0).astype('int64')  # P
            labels = labels.numpy().astype('int64')  # T
            evaluate_worker(self, labels, preds)
        elif isinstance(preds, (list, tuple)):
            threads = [threading.Thread(target=evaluate_worker,
                                        args=(self, label, pred),
                                        )
                       for (label, pred) in zip(labels, preds)]
            for thread in threads:
                thread.start()
            for thread in threads:
                thread.join()
        #elif preds.dtype == numpy.uint8:
        elif isinstance(preds, np.ndarray):
            preds = ((preds / np.max(preds)) > 0.5).astype('int64')  # P
            labels = (labels / np.max(labels)).astype('int64')  # T
            evaluate_worker(self, labels, preds)
        else:
            raise NotImplemented

    def get_all(self):
        return self.total_tp, self.total_fp, self.total_fn

    def get(self):
        return get_miou_prec_recall_fscore(self.total_tp, self.total_fp, self.total_fn)

    def reset(self):
        self.total_tp = 0
        self.total_fp = 0
        self.total_fn = 0
        return

def batch_tp_fp_fn(predict, target, nclass):
    """Batch Intersection of Union
    Args:
        predict: input 4D tensor
        target: label 3D tensor
        nclass: number of categories (int)
    """

    mini = 1
    maxi = nclass
    nbins = nclass

    # predict = (output.detach().numpy() > 0).astype('int64')  # P
    # target = target.numpy().astype('int64')  # T
    intersection = predict * (predict == target)  # TP

    # areas of intersection and union
    area_inter, _ = np.histogram(intersection, bins=nbins, range=(mini, maxi))
    area_pred, _ = np.histogram(predict, bins=nbins, range=(mini, maxi))
    area_lab, _ = np.histogram(target, bins=nbins, range=(mini, maxi))

    # areas of TN FP FN
    area_tp = area_inter[0]
    area_fp = area_pred[0] - area_inter[0]
    area_fn = area_lab[0] - area_inter[0]

    # area_union = area_pred + area_lab - area_inter
    assert area_tp <= (area_tp + area_fn + area_fp)
    return area_tp, area_fp, area_fn




"""
工具包，主要包含Pepper项目中一些深度学习常用的操作
"""

from .environment import cv2, torch, plt, np

"""
1.测试工具
概述：用于检测python包
作用：打印print测试Pepper安装效果
"""
def printPepper():
    print("welcome to Pepper!")




'''
2.Document_scanning
概况：文档扫描功能
'''
def Document_scanning(img):
    img_gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    # OTSU
    ret, thresh_image = cv2.threshold(img_gray, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)

    # 高斯模糊
    binary = cv2.GaussianBlur(img_gray, (5, 5), 0)

    # 边缘检测
    binary = cv2.Canny(binary, int(ret / 3), ret, apertureSize=3)

    # 膨胀操作，尽量使边缘闭合
    kernel = np.ones((3, 3), np.uint8)
    img_binary = cv2.dilate(binary, kernel, iterations=1)

    # 寻找边缘
    contours, _ = cv2.findContours(img_binary, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
    # 计算面积
    max_area = 0.0
    max_contour = []
    for contour in contours:
        currentArea = cv2.contourArea(contour)
        if currentArea > max_area:
            max_area = currentArea
            max_contour = contour

    # 多边形拟合凸包
    hull = cv2.convexHull(max_contour)
    epsilon = 0.02 * cv2.arcLength(max_contour, True)
    approx = cv2.approxPolyDP(hull, epsilon, True)
    boxes = approx.reshape((len(approx), 2))

    rect = np.zeros((4, 2), dtype='float32')
    s = boxes.sum(axis=1)
    rect[0] = boxes[np.argmin(s)]
    rect[2] = boxes[np.argmax(s)]
    diff = np.diff(boxes, axis=1)
    rect[1] = boxes[np.argmin(diff)]
    rect[3] = boxes[np.argmax(diff)]
    (tl, tr, br, bl) = rect

    # 计算输入的w和h的值
    widthA = np.sqrt(((br[0] - bl[0]) ** 2) + ((br[1] - bl[1]) ** 2))
    widthB = np.sqrt(((tr[0] - tl[0]) ** 2) + ((tr[1] - tl[1]) ** 2))
    maxWidth = max(int(widthA), int(widthB))
    heightA = np.sqrt(((tr[0] - br[0]) ** 2) + ((tr[1] - br[1]) ** 2))
    heightB = np.sqrt(((tl[0] - bl[0]) ** 2) + ((tl[1] - bl[1]) ** 2))
    maxHeight = max(int(heightA), int(heightB))
    # 变化后对应坐标位置
    dst = np.array([
        [0, 0],
        [maxWidth - 1, 0],
        [maxWidth - 1, maxHeight - 1],
        [0, maxHeight - 1]],
        dtype='float32')

    # 计算变换矩阵
    M = cv2.getPerspectiveTransform(rect, dst)
    warped = cv2.warpPerspective(img, M, (maxWidth, maxHeight))

    return warped

































# import torch
# import torch.nn as nn
# import torch.nn.functional as F
# from utils import *
from PepperPepper.environment import torch, nn, F
from PepperPepper.IRSTD.tools.utils import *


class SoftIoULoss(nn.Module):
    def __init__(self):
        super(SoftIoULoss, self).__init__()
    def forward(self, preds, gt_masks):
        if isinstance(preds, list) or isinstance(preds, tuple):
            loss_total = 0
            for i in range(len(preds)):
                pred = preds[i]
                smooth = 1
                intersection = pred * gt_masks
                loss = (intersection.sum() + smooth) / (pred.sum() + gt_masks.sum() -intersection.sum() + smooth)
                loss = 1 - loss.mean()
                loss_total = loss_total + loss
            return loss_total / len(preds)
        else:
            pred = preds
            smooth = 1
            intersection = pred * gt_masks
            loss = (intersection.sum() + smooth) / (pred.sum() + gt_masks.sum() -intersection.sum() + smooth)
            loss = 1 - loss.mean()
            return loss

class ISNetLoss(nn.Module):
    def __init__(self):
        super(ISNetLoss, self).__init__()
        self.softiou = SoftIoULoss()
        self.bce = nn.BCELoss()
        self.grad = Get_gradient_nopadding()
        
    def forward(self, preds, gt_masks):
        edge_gt = self.grad(gt_masks.clone())
        
        ### img loss
        loss_img = self.softiou(preds[0], gt_masks)
        
        ### edge loss
        loss_edge = 10 * self.bce(preds[1], edge_gt)+ self.softiou(preds[1].sigmoid(), edge_gt)
        
        return loss_img + loss_edge



import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange, repeat
from timm.models.layers import DropPath, to_2tuple, trunc_normal_
try:
    from mamba_ssm.ops.selective_scan_interface import selective_scan_fn, selective_scan_ref
except:
    pass

# an alternative for mamba_ssm (in which causal_conv1d is needed)
try:
    from selective_scan import selective_scan_fn as selective_scan_fn_v1
    from selective_scan import selective_scan_ref as selective_scan_ref_v1
except:
    pass

from PepperPepper.environment import profile




class SS2D(nn.Module):
    def __init__(
        self,
        d_model,
        d_state=16,
        # d_state="auto", # 20240109
        d_conv=3,
        expand=2,
        dt_rank="auto",
        dt_min=0.001,
        dt_max=0.1,
        dt_init="random",
        dt_scale=1.0,
        dt_init_floor=1e-4,
        dropout=0.,
        conv_bias=True,
        bias=False,
        device=None,
        dtype=None,
        **kwargs,
    ):
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.d_model = d_model
        self.d_state = d_state
        # self.d_state = math.ceil(self.d_model / 6) if d_state == "auto" else d_model # 20240109
        self.d_conv = d_conv
        self.expand = expand
        self.d_inner = int(self.expand * self.d_model)
        self.dt_rank = math.ceil(self.d_model / 16) if dt_rank == "auto" else dt_rank

        self.in_proj = nn.Linear(self.d_model, self.d_inner * 2, bias=bias, **factory_kwargs)
        self.conv2d = nn.Conv2d(
            in_channels=self.d_inner,
            out_channels=self.d_inner,
            groups=self.d_inner,
            bias=conv_bias,
            kernel_size=d_conv,
            padding=(d_conv - 1) // 2,
            **factory_kwargs,
        )
        self.act = nn.SiLU()

        self.x_proj = (
            nn.Linear(self.d_inner, (self.dt_rank + self.d_state * 2), bias=False, **factory_kwargs), 
            nn.Linear(self.d_inner, (self.dt_rank + self.d_state * 2), bias=False, **factory_kwargs), 
            nn.Linear(self.d_inner, (self.dt_rank + self.d_state * 2), bias=False, **factory_kwargs), 
            nn.Linear(self.d_inner, (self.dt_rank + self.d_state * 2), bias=False, **factory_kwargs), 
        )
        self.x_proj_weight = nn.Parameter(torch.stack([t.weight for t in self.x_proj], dim=0)) # (K=4, N, inner)
        del self.x_proj

        self.dt_projs = (
            self.dt_init(self.dt_rank, self.d_inner, dt_scale, dt_init, dt_min, dt_max, dt_init_floor, **factory_kwargs),
            self.dt_init(self.dt_rank, self.d_inner, dt_scale, dt_init, dt_min, dt_max, dt_init_floor, **factory_kwargs),
            self.dt_init(self.dt_rank, self.d_inner, dt_scale, dt_init, dt_min, dt_max, dt_init_floor, **factory_kwargs),
            self.dt_init(self.dt_rank, self.d_inner, dt_scale, dt_init, dt_min, dt_max, dt_init_floor, **factory_kwargs),
        )
        self.dt_projs_weight = nn.Parameter(torch.stack([t.weight for t in self.dt_projs], dim=0)) # (K=4, inner, rank)
        self.dt_projs_bias = nn.Parameter(torch.stack([t.bias for t in self.dt_projs], dim=0)) # (K=4, inner)
        del self.dt_projs
        
        self.A_logs = self.A_log_init(self.d_state, self.d_inner, copies=4, merge=True) # (K=4, D, N)
        self.Ds = self.D_init(self.d_inner, copies=4, merge=True) # (K=4, D, N)

        # self.selective_scan = selective_scan_fn
        self.forward_core = self.forward_corev0

        self.out_norm = nn.LayerNorm(self.d_inner)
        self.out_proj = nn.Linear(self.d_inner, self.d_model, bias=bias, **factory_kwargs)
        self.dropout = nn.Dropout(dropout) if dropout > 0. else None

    @staticmethod
    def dt_init(dt_rank, d_inner, dt_scale=1.0, dt_init="random", dt_min=0.001, dt_max=0.1, dt_init_floor=1e-4, **factory_kwargs):
        dt_proj = nn.Linear(dt_rank, d_inner, bias=True, **factory_kwargs)

        # Initialize special dt projection to preserve variance at initialization
        dt_init_std = dt_rank**-0.5 * dt_scale
        if dt_init == "constant":
            nn.init.constant_(dt_proj.weight, dt_init_std)
        elif dt_init == "random":
            nn.init.uniform_(dt_proj.weight, -dt_init_std, dt_init_std)
        else:
            raise NotImplementedError

        # Initialize dt bias so that F.softplus(dt_bias) is between dt_min and dt_max
        dt = torch.exp(
            torch.rand(d_inner, **factory_kwargs) * (math.log(dt_max) - math.log(dt_min))
            + math.log(dt_min)
        ).clamp(min=dt_init_floor)
        # Inverse of softplus: https://github.com/pytorch/pytorch/issues/72759
        inv_dt = dt + torch.log(-torch.expm1(-dt))
        with torch.no_grad():
            dt_proj.bias.copy_(inv_dt)
        # Our initialization would set all Linear.bias to zero, need to mark this one as _no_reinit
        dt_proj.bias._no_reinit = True
        
        return dt_proj

    @staticmethod
    def A_log_init(d_state, d_inner, copies=1, device=None, merge=True):
        # S4D real initialization
        A = repeat(
            torch.arange(1, d_state + 1, dtype=torch.float32, device=device),
            "n -> d n",
            d=d_inner,
        ).contiguous()
        A_log = torch.log(A)  # Keep A_log in fp32
        if copies > 1:
            A_log = repeat(A_log, "d n -> r d n", r=copies)
            if merge:
                A_log = A_log.flatten(0, 1)
        A_log = nn.Parameter(A_log)
        A_log._no_weight_decay = True
        return A_log

    @staticmethod
    def D_init(d_inner, copies=1, device=None, merge=True):
        # D "skip" parameter
        D = torch.ones(d_inner, device=device)
        if copies > 1:
            D = repeat(D, "n1 -> r n1", r=copies)
            if merge:
                D = D.flatten(0, 1)
        D = nn.Parameter(D)  # Keep in fp32
        D._no_weight_decay = True
        return D

    def forward_corev0(self, x: torch.Tensor):
        self.selective_scan = selective_scan_fn
        
        B, C, H, W = x.shape
        L = H * W
        K = 4

        x_hwwh = torch.stack([x.view(B, -1, L), torch.transpose(x, dim0=2, dim1=3).contiguous().view(B, -1, L)], dim=1).view(B, 2, -1, L)
        xs = torch.cat([x_hwwh, torch.flip(x_hwwh, dims=[-1])], dim=1) # (b, k, d, l)

        x_dbl = torch.einsum("b k d l, k c d -> b k c l", xs.view(B, K, -1, L), self.x_proj_weight)
        # x_dbl = x_dbl + self.x_proj_bias.view(1, K, -1, 1)
        dts, Bs, Cs = torch.split(x_dbl, [self.dt_rank, self.d_state, self.d_state], dim=2)
        dts = torch.einsum("b k r l, k d r -> b k d l", dts.view(B, K, -1, L), self.dt_projs_weight)
        # dts = dts + self.dt_projs_bias.view(1, K, -1, 1)

        xs = xs.float().view(B, -1, L) # (b, k * d, l)
        dts = dts.contiguous().float().view(B, -1, L) # (b, k * d, l)
        Bs = Bs.float().view(B, K, -1, L) # (b, k, d_state, l)
        Cs = Cs.float().view(B, K, -1, L) # (b, k, d_state, l)
        Ds = self.Ds.float().view(-1) # (k * d)
        As = -torch.exp(self.A_logs.float()).view(-1, self.d_state)  # (k * d, d_state)
        dt_projs_bias = self.dt_projs_bias.float().view(-1) # (k * d)

        out_y = self.selective_scan(
            xs, dts, 
            As, Bs, Cs, Ds, z=None,
            delta_bias=dt_projs_bias,
            delta_softplus=True,
            return_last_state=False,
        ).view(B, K, -1, L)
        assert out_y.dtype == torch.float

        inv_y = torch.flip(out_y[:, 2:4], dims=[-1]).view(B, 2, -1, L)
        wh_y = torch.transpose(out_y[:, 1].view(B, -1, W, H), dim0=2, dim1=3).contiguous().view(B, -1, L)
        invwh_y = torch.transpose(inv_y[:, 1].view(B, -1, W, H), dim0=2, dim1=3).contiguous().view(B, -1, L)

        return out_y[:, 0], inv_y[:, 0], wh_y, invwh_y

    # an alternative to forward_corev1
    def forward_corev1(self, x: torch.Tensor):
        self.selective_scan = selective_scan_fn_v1

        B, C, H, W = x.shape
        L = H * W
        K = 4

        x_hwwh = torch.stack([x.view(B, -1, L), torch.transpose(x, dim0=2, dim1=3).contiguous().view(B, -1, L)], dim=1).view(B, 2, -1, L)
        xs = torch.cat([x_hwwh, torch.flip(x_hwwh, dims=[-1])], dim=1) # (b, k, d, l)

        x_dbl = torch.einsum("b k d l, k c d -> b k c l", xs.view(B, K, -1, L), self.x_proj_weight)
        # x_dbl = x_dbl + self.x_proj_bias.view(1, K, -1, 1)
        dts, Bs, Cs = torch.split(x_dbl, [self.dt_rank, self.d_state, self.d_state], dim=2)
        dts = torch.einsum("b k r l, k d r -> b k d l", dts.view(B, K, -1, L), self.dt_projs_weight)
        # dts = dts + self.dt_projs_bias.view(1, K, -1, 1)

        xs = xs.float().view(B, -1, L) # (b, k * d, l)
        dts = dts.contiguous().float().view(B, -1, L) # (b, k * d, l)
        Bs = Bs.float().view(B, K, -1, L) # (b, k, d_state, l)
        Cs = Cs.float().view(B, K, -1, L) # (b, k, d_state, l)
        Ds = self.Ds.float().view(-1) # (k * d)
        As = -torch.exp(self.A_logs.float()).view(-1, self.d_state)  # (k * d, d_state)
        dt_projs_bias = self.dt_projs_bias.float().view(-1) # (k * d)

        out_y = self.selective_scan(
            xs, dts, 
            As, Bs, Cs, Ds,
            delta_bias=dt_projs_bias,
            delta_softplus=True,
        ).view(B, K, -1, L)
        assert out_y.dtype == torch.float

        inv_y = torch.flip(out_y[:, 2:4], dims=[-1]).view(B, 2, -1, L)
        wh_y = torch.transpose(out_y[:, 1].view(B, -1, W, H), dim0=2, dim1=3).contiguous().view(B, -1, L)
        invwh_y = torch.transpose(inv_y[:, 1].view(B, -1, W, H), dim0=2, dim1=3).contiguous().view(B, -1, L)

        return out_y[:, 0], inv_y[:, 0], wh_y, invwh_y

    def forward(self, x, H, W, relative_pos=None):
        B, N, C = x.shape
        #print('x input',x.shape)
        x = x.permute(0, 2, 1).reshape(B, H, W, C)

        B, H, W, C = x.shape

        xz = self.in_proj(x)
        x, z = xz.chunk(2, dim=-1) # (b, h, w, d)

        x = x.permute(0, 3, 1, 2).contiguous()
        x = self.act(self.conv2d(x)) # (b, d, h, w)
        y1, y2, y3, y4 = self.forward_core(x)
        assert y1.dtype == torch.float32
        y = y1 + y2 + y3 + y4
        y = torch.transpose(y, dim0=1, dim1=2).contiguous().view(B, H, W, -1)
        y = self.out_norm(y)
        y = y * F.silu(z)
        out = self.out_proj(y)
        if self.dropout is not None:
            out = self.dropout(out)
        out=out.reshape(B,N,C)
        #print('x output',out.shape)
        return out

class Mlp(nn.Module):
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x

def make_pairs(x):
    """make the int -> tuple 
    """
    return x if isinstance(x, tuple) else (x, x)

class InvertedResidualFeedForward(nn.Module):
    def __init__(self, dim, dim_ratio=2.):
        super(InvertedResidualFeedForward, self).__init__()
        output_dim = int(dim_ratio * dim)
        self.conv1x1_gelu_bn = ConvGeluBN(
            in_channel=dim,
            out_channel=output_dim,
            kernel_size=1,
            stride_size=1,
            padding=0
        )
        self.conv3x3_dw = ConvDW3x3(dim=output_dim)  
        self.act = nn.Sequential(
            nn.GELU(),
            nn.BatchNorm2d(output_dim)
        )
        self.conv1x1_pw = nn.Sequential(
            nn.Conv2d(output_dim, dim, 1, 1, 0),
            nn.BatchNorm2d(dim)
        )
        
    def forward(self, x):
        x = self.conv1x1_gelu_bn(x)
        out = x + self.act(self.conv3x3_dw(x))
        out = self.conv1x1_pw(out)
        return out 


class ConvDW3x3(nn.Module):
    def __init__(self, dim, kernel_size=3):
        super(ConvDW3x3, self).__init__()
        self.conv = nn.Conv2d(
            in_channels=dim, 
            out_channels=dim, 
            kernel_size=make_pairs(kernel_size),
            padding=make_pairs(1),
            groups=dim)
    
    def forward(self, x):
        x = self.conv(x)
        return x 


class ConvGeluBN(nn.Module):
    def __init__(self, in_channel, out_channel, kernel_size, stride_size, padding=1):
        """build the conv3x3 + gelu + bn module
        """
        super(ConvGeluBN, self).__init__()
        self.kernel_size = make_pairs(kernel_size)
        self.stride_size = make_pairs(stride_size)
        self.padding_size = make_pairs(padding)
        self.in_channel = in_channel
        self.out_channel = out_channel 
        self.conv3x3_gelu_bn = nn.Sequential(
            nn.Conv2d(in_channels=self.in_channel,
                      out_channels=self.out_channel,
                      kernel_size=self.kernel_size,
                      stride=self.stride_size,
                      padding=self.padding_size),
            nn.GELU(),
            nn.BatchNorm2d(self.out_channel)
        )

    def forward(self, x):
        x = self.conv3x3_gelu_bn(x)
        return x 

class Block(nn.Module):
    """ MiM-ISTD Block
    """
    def __init__(self, outer_dim, inner_dim, outer_head, inner_head, num_words, mlp_ratio=4.,
                 qkv_bias=False, qk_scale=None, drop=0., attn_drop=0., drop_path=0., act_layer=nn.GELU,
                 norm_layer=nn.LayerNorm, se=0, sr_ratio=1):
        super().__init__()
        self.has_inner = inner_dim > 0
        if self.has_inner:
            # Inner
            self.inner_norm1 = norm_layer(num_words * inner_dim)
            self.inner_attn = SS2D(d_model=inner_dim, dropout=0, d_state=16)
            self.inner_norm2 = norm_layer(num_words * inner_dim)
            self.inner_mlp = InvertedResidualFeedForward(inner_dim)
            # self.inner_mlp = Mlp(in_features=inner_dim, hidden_features=int(inner_dim * mlp_ratio),
            #                      out_features=inner_dim, act_layer=act_layer, drop=drop)

            self.proj_norm1 = norm_layer(num_words * inner_dim)
            self.proj = nn.Linear(num_words * inner_dim, outer_dim, bias=False)
            self.proj_norm2 = norm_layer(outer_dim)
        # Outer
        self.outer_norm1 = norm_layer(outer_dim)

        self.outer_attn = SS2D(d_model=outer_dim, dropout=0, d_state=16)
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.outer_norm2 = norm_layer(outer_dim)
        self.outer_mlp = InvertedResidualFeedForward(outer_dim)
        # self.outer_mlp = Mlp(in_features=outer_dim, hidden_features=int(outer_dim * mlp_ratio),
        #                      out_features=outer_dim, act_layer=act_layer, drop=drop)

    def forward(self, x, outer_tokens, H_out, W_out, H_in, W_in, relative_pos):
        B, N, C = outer_tokens.size()
        #print('outer_tokens input',outer_tokens.shape)
        if self.has_inner:
            x = x + self.drop_path(self.inner_attn(self.inner_norm1(x.reshape(B, N, -1)).reshape(B*N, H_in*W_in, -1), H_in, W_in)) # B*N, k*k, c
            mid=self.inner_norm2(x.reshape(B, N, -1)).reshape(B*N, H_in*W_in, -1)
            mid=mid.reshape(B,mid.size(-1),int(math.sqrt(N*H_in*W_in)),int(math.sqrt(N*H_in*W_in)))
            x = x + self.drop_path(self.inner_mlp(mid).reshape(B*N, H_in*W_in, -1)).reshape(B*N, H_in*W_in, -1)
            #x = x + self.drop_path(self.inner_mlp(self.inner_norm2(x.reshape(B, N, -1)).reshape(B*N, H_in*W_in, -1))) # B*N, k*k, c
            outer_tokens = outer_tokens + self.proj_norm2(self.proj(self.proj_norm1(x.reshape(B, N, -1)))) # B, N, C
        outer_tokens = outer_tokens + self.drop_path(self.outer_attn(self.outer_norm1(outer_tokens), H_out, W_out, relative_pos))
        mid_out=self.outer_norm2(outer_tokens)
        mid_out=mid_out.reshape(B,mid_out.size(-1),int(math.sqrt(N)),int(math.sqrt(N)))
        outer_tokens = outer_tokens + self.drop_path(self.outer_mlp(mid_out).reshape(B,N,C))
        return x, outer_tokens






class PatchMerging2D_sentence(nn.Module):
    r""" Patch Merging Layer.
    Args:
        input_resolution (tuple[int]): Resolution of input feature.
        dim (int): Number of input channels.
        norm_layer (nn.Module, optional): Normalization layer.  Default: nn.LayerNorm
    """

    def __init__(self, dim, norm_layer=nn.LayerNorm):
        super().__init__()
        self.dim = dim
        self.reduction = nn.Linear(4 * dim, 2 * dim, bias=False)
        self.norm = norm_layer(4 * dim)

    def forward(self, x):#(b,h,w,c)->(b,h/2,w/2,2c)
        B, N, C = x.shape
        x=x.reshape(B,int(math.sqrt(N)),int(math.sqrt(N)),C)
        B, H, W, C = x.shape

        SHAPE_FIX = [-1, -1]
        if (W % 2 != 0) or (H % 2 != 0):
            print(f"Warning, x.shape {x.shape} is not match even ===========", flush=True)
            SHAPE_FIX[0] = H // 2
            SHAPE_FIX[1] = W // 2

        x0 = x[:, 0::2, 0::2, :]  # B H/2 W/2 C
        x1 = x[:, 1::2, 0::2, :]  # B H/2 W/2 C
        x2 = x[:, 0::2, 1::2, :]  # B H/2 W/2 C
        x3 = x[:, 1::2, 1::2, :]  # B H/2 W/2 C

        if SHAPE_FIX[0] > 0:
            x0 = x0[:, :SHAPE_FIX[0], :SHAPE_FIX[1], :]
            x1 = x1[:, :SHAPE_FIX[0], :SHAPE_FIX[1], :]
            x2 = x2[:, :SHAPE_FIX[0], :SHAPE_FIX[1], :]
            x3 = x3[:, :SHAPE_FIX[0], :SHAPE_FIX[1], :]
        
        x = torch.cat([x0, x1, x2, x3], -1)  # B H/2 W/2 4*C
        x = x.view(B, H//2, W//2, 4 * C)  # B H/2*W/2 4*C

        x = self.norm(x)
        x = self.reduction(x)
        b, h, w, c = x.shape
        x=x.reshape(b,h*w,c)

        return x,h,w


class PatchMerging2D_word(nn.Module):
    r""" Patch Merging Layer.
    Args:
        input_resolution (tuple[int]): Resolution of input feature.
        dim (int): Number of input channels.
        norm_layer (nn.Module, optional): Normalization layer.  Default: nn.LayerNorm
    """

    def __init__(self, dim_in, dim_out, stride=2, act_layer=nn.GELU):
        super().__init__()
        self.stride = stride
        self.dim_out = dim_out
        self.norm = nn.LayerNorm(dim_in)
        self.conv = nn.Sequential(
            nn.Conv2d(dim_in, dim_out, kernel_size=2*stride-1, padding=stride-1, stride=stride),
        )

    def forward(self, x, H_out, W_out, H_in, W_in):
        B_N, M, C = x.shape # B*N, M, C
        x = self.norm(x)
        x = x.reshape(-1, H_out, W_out, H_in, W_in, C)
        # padding to fit (1333, 800) in detection.
        pad_input = (H_out % 2 == 1) or (W_out % 2 == 1)
        if pad_input:
            x = F.pad(x.permute(0, 3, 4, 5, 1, 2), (0, W_out % 2, 0, H_out % 2))
            x = x.permute(0, 4, 5, 1, 2, 3)

        H,W=x.shape[1],x.shape[2]      
        SHAPE_FIX = [-1, -1]
        if (W % 2 != 0) or (H % 2 != 0):
            print(f"Warning, x.shape {x.shape} is not match even ===========", flush=True)
            SHAPE_FIX[0] = H // 2
            SHAPE_FIX[1] = W // 2

        x0 = x[:, 0::2, 0::2, :]  # B H/2 W/2 C
        x1 = x[:, 1::2, 0::2, :]  # B H/2 W/2 C
        x2 = x[:, 0::2, 1::2, :]  # B H/2 W/2 C
        x3 = x[:, 1::2, 1::2, :]  # B H/2 W/2 C

        if SHAPE_FIX[0] > 0:
            x0 = x0[:, :SHAPE_FIX[0], :SHAPE_FIX[1], :]
            x1 = x1[:, :SHAPE_FIX[0], :SHAPE_FIX[1], :]
            x2 = x2[:, :SHAPE_FIX[0], :SHAPE_FIX[1], :]
            x3 = x3[:, :SHAPE_FIX[0], :SHAPE_FIX[1], :]

        x = torch.cat([torch.cat([x0, x1], 3), torch.cat([x2, x3], 3)], 4) # B, H/2, W/2, 2*H_in, 2*W_in, C
        x = x.reshape(-1, 2*H_in, 2*W_in, C).permute(0, 3, 1, 2) # B_N/4, C, 2*H_in, 2*W_in
        x = self.conv(x)  # B_N/4, C, H_in, W_in
        x = x.reshape(-1, self.dim_out, M).transpose(1, 2)

        return x
    



class Stem(nn.Module):

    def __init__(self, img_size=224, in_chans=3, outer_dim=768, inner_dim=24):
        super().__init__()
        img_size = to_2tuple(img_size)
        self.img_size = img_size
        self.inner_dim = inner_dim
        self.num_patches = img_size[0] // 8 * img_size[1] // 8
        self.num_words = 16
        
        self.common_conv = nn.Sequential(
            nn.Conv2d(in_chans, inner_dim*2, 3, stride=2, padding=1),
            nn.BatchNorm2d(inner_dim*2),
            nn.ReLU(inplace=True),
        )
        self.inner_convs = nn.Sequential(
            nn.Conv2d(inner_dim*2, inner_dim, 3, stride=1, padding=1),
            nn.BatchNorm2d(inner_dim),
            nn.ReLU(inplace=False),
        )
        self.outer_convs = nn.Sequential(
            nn.Conv2d(inner_dim*2, inner_dim*4, 3, stride=2, padding=1),
            nn.BatchNorm2d(inner_dim*4),
            nn.ReLU(inplace=True),
            nn.Conv2d(inner_dim*4, inner_dim*8, 3, stride=2, padding=1),
            nn.BatchNorm2d(inner_dim*8),
            nn.ReLU(inplace=True),
            nn.Conv2d(inner_dim*8, outer_dim, 3, stride=1, padding=1),
            nn.BatchNorm2d(outer_dim),
            nn.ReLU(inplace=False),
        )
        self.unfold = nn.Unfold(kernel_size=4, padding=0, stride=4)
        
    def forward(self, x):
        B, C, H, W = x.shape
        x = self.common_conv(x)     
        H_out, W_out = H // 8, W // 8  # Each visual sentence corresponds to 8x8 pixel area of the original image
        H_in, W_in = 4, 4 # Every visual sentence is composed of 4x4 visual words, Every visual word at the stem stage corresponds to 2x2 pixel area of the original image
        # inner_tokens
        inner_tokens = self.inner_convs(x) # B, C, H, W
        inner_tokens = self.unfold(inner_tokens).transpose(1, 2) # B, N, Ck2
        inner_tokens = inner_tokens.reshape(B * H_out * W_out, self.inner_dim, H_in*W_in).transpose(1, 2) # B*N, C, 4*4
        # outer_tokens
        outer_tokens = self.outer_convs(x) # B, C, H_out, W_out
        outer_tokens = outer_tokens.permute(0, 2, 3, 1).reshape(B, H_out * W_out, -1)
        return inner_tokens, outer_tokens, (H_out, W_out), (H_in, W_in)

class Stage(nn.Module):
    """ PyramidTNT stage
    """
    def __init__(self, num_blocks, outer_dim, inner_dim, outer_head, inner_head, num_patches, num_words, mlp_ratio=4.,
                 qkv_bias=False, qk_scale=None, drop=0., attn_drop=0., drop_path=0., act_layer=nn.GELU,
                 norm_layer=nn.LayerNorm, se=0, sr_ratio=1):
        super().__init__()
        blocks = []
        drop_path = drop_path if isinstance(drop_path, list) else [drop_path] * num_blocks
        
        for j in range(num_blocks):
            if j == 0:
                _inner_dim = inner_dim
            elif j == 1 and num_blocks > 6:
                _inner_dim = inner_dim
            else:
                _inner_dim = -1
            blocks.append(Block(
                outer_dim, _inner_dim, outer_head=outer_head, inner_head=inner_head,
                num_words=num_words, mlp_ratio=mlp_ratio, qkv_bias=qkv_bias, qk_scale=qk_scale, drop=drop,
                attn_drop=attn_drop, drop_path=drop_path[j], act_layer=act_layer, norm_layer=norm_layer,
                se=se, sr_ratio=sr_ratio))

        self.blocks = nn.ModuleList(blocks)
        self.relative_pos = nn.Parameter(torch.randn(
                        1, outer_head, num_patches, num_patches // sr_ratio // sr_ratio))

    def forward(self, inner_tokens, outer_tokens, H_out, W_out, H_in, W_in):
        for blk in self.blocks:
            inner_tokens, outer_tokens = blk(inner_tokens, outer_tokens, H_out, W_out, H_in, W_in, self.relative_pos)
        return inner_tokens, outer_tokens


class UpsampleBlock(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(UpsampleBlock, self).__init__()
        # 步长为2的2x2转置卷积
        self.transposed_conv = nn.ConvTranspose2d(
            in_channels, out_channels, kernel_size=2, stride=2, padding=0
        )
        # 批量归一化
        self.batch_norm1 = nn.BatchNorm2d(out_channels)
        # GeLU 激活函数
        self.gelu1 = nn.GELU()
        # 步长为1的3x3卷积
        self.conv = nn.Conv2d(
            out_channels, out_channels, kernel_size=3, stride=1, padding=1
        )
        # 另一个批量归一化
        self.batch_norm2 = nn.BatchNorm2d(out_channels)
        # 另一个 GeLU 激活函数
        self.gelu2 = nn.GELU()

    def forward(self, x):
        x = self.transposed_conv(x)
        x = self.batch_norm1(x)
        x = self.gelu1(x)
        x = self.conv(x)
        x = self.batch_norm2(x)
        x = self.gelu2(x)
        return x




class PyramidMiM_enc(nn.Module):
    """ Pyramid MiM-ISTD encoder including conv stem for computer vision
    """
    def __init__(self, configs=None, img_size=256, in_chans=1, num_classes=1, mlp_ratio=4., qkv_bias=False,
                qk_scale=None, drop_rate=0., attn_drop_rate=0., drop_path_rate=0., norm_layer=nn.LayerNorm, se=0):
        super().__init__()
        self.num_classes = num_classes
        depths = [2, 2, 2, 2]
        outer_dims = [32, 32*2, 32*4, 32*8]
        inner_dims = [4, 4*2, 4*4, 4*8]#  original mim-istd
        outer_heads = [2, 2*2, 2*4, 2*8]
        inner_heads = [1, 1*2, 1*4, 1*8]
        sr_ratios = [4, 2, 1, 1]
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))] 
        self.num_features = outer_dims[-1]     

        self.patch_embed = Stem(
            img_size=img_size, in_chans=in_chans, outer_dim=outer_dims[0], inner_dim=inner_dims[0])
        num_patches = self.patch_embed.num_patches
        num_words = self.patch_embed.num_words
        

        depth = 0
        self.word_merges = nn.ModuleList([])
        self.sentence_merges = nn.ModuleList([])
        self.stages = nn.ModuleList([])
        for i in range(4):
            if i > 0:
                self.word_merges.append(PatchMerging2D_word(inner_dims[i-1], inner_dims[i]))
                self.sentence_merges.append(PatchMerging2D_sentence(outer_dims[i-1]))
            self.stages.append(Stage(depths[i], outer_dim=outer_dims[i], inner_dim=inner_dims[i],
                        outer_head=outer_heads[i], inner_head=inner_heads[i],
                        num_patches=num_patches // (2 ** i) // (2 ** i), num_words=num_words, mlp_ratio=mlp_ratio,
                        qkv_bias=qkv_bias, qk_scale=qk_scale, drop=drop_rate, attn_drop=attn_drop_rate,
                        drop_path=dpr[depth:depth+depths[i]], norm_layer=norm_layer, se=se, sr_ratio=sr_ratios[i])
            )
            depth += depths[i]
        
        self.norm = norm_layer(outer_dims[-1])

        self.up_blocks = nn.ModuleList([])
        for i in range(4):
            self.up_blocks.append(UpsampleBlock(outer_dims[i],outer_dims[i]))
           
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        if isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    @torch.jit.ignore
    def no_weight_decay(self):
        return {'outer_pos', 'inner_pos'}

    def get_classifier(self):
        return self.head

    def reset_classifier(self, num_classes, global_pool=''):
        self.num_classes = num_classes
        self.head = nn.Linear(self.num_features, num_classes) if num_classes > 0 else nn.Identity()

    def forward_features(self, x):
        size = x.size()[2:]
        inner_tokens, outer_tokens, (H_out, W_out), (H_in, W_in) = self.patch_embed(x)
        outputs=[]
        
        for i in range(4):
            if i > 0:
                inner_tokens = self.word_merges[i-1](inner_tokens, H_out, W_out, H_in, W_in)
                outer_tokens, H_out, W_out = self.sentence_merges[i-1](outer_tokens)
            inner_tokens, outer_tokens = self.stages[i](inner_tokens, outer_tokens, H_out, W_out, H_in, W_in)
            b,l,m=outer_tokens.shape
            mid_out=outer_tokens.reshape(b,int(math.sqrt(l)),int(math.sqrt(l)),m).permute(0,3,1,2)
            mid_out=self.up_blocks[i](mid_out)

            outputs.append(mid_out)

        return outputs

    def forward(self, x):
        x = self.forward_features(x)
        return x


class ResidualBlock(nn.Module):
    def __init__(self, in_channels, out_channels, stride, downsample):
        super(ResidualBlock, self).__init__()
        self.body = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, 3, stride, 1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(True),

            nn.Conv2d(out_channels, out_channels, 3, 1, 1, bias=False),
            nn.BatchNorm2d(out_channels),
        )
        if downsample:
            self.downsample = nn.Sequential(
                nn.Conv2d(in_channels, out_channels, 1, stride, 0, bias=False),
                nn.BatchNorm2d(out_channels),
            )
        else:
            self.downsample = nn.Sequential()


    def forward(self, x):
        residual = x
        x = self.body(x)

        if self.downsample:
            residual = self.downsample(residual)

        out = F.relu(x+residual, True)
        return out

class _FCNHead(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(_FCNHead, self).__init__()
        inter_channels = in_channels // 4
        self.block = nn.Sequential(
            nn.Conv2d(in_channels, inter_channels, 3, 1, 1, bias=False),
            nn.BatchNorm2d(inter_channels),
            nn.ReLU(True),
            nn.Dropout(0.1),
            nn.Conv2d(inter_channels, out_channels, 1, 1, 0)
        )

    def forward(self, x):
        return self.block(x)

class PatchExpand2D(nn.Module):
    def __init__(self, dim, dim_scale=2, norm_layer=nn.LayerNorm):
        super().__init__()
        self.dim = dim*2
        self.dim_scale = dim_scale
        self.expand = nn.Linear(self.dim, dim_scale*self.dim, bias=False)
        self.norm = norm_layer(self.dim // dim_scale)

    def forward(self, x):#(b,h,w,c)->(b,h,w,2c)->(b,2h,2w,c/2)
        x=x.permute(0,2,3,1)
        B, H, W, C = x.shape
        x = self.expand(x)

        x = rearrange(x, 'b h w (p1 p2 c)-> b (h p1) (w p2) c', p1=self.dim_scale, p2=self.dim_scale, c=C//self.dim_scale)
        x= self.norm(x).permute(0,3,1,2)

        return x

    # [2] * 3, [8, 16, 32, 64, 128]


class MiM(nn.Module): 
    def __init__(self, layer_blocks= [2] * 3, channels = [8, 16, 32, 64, 128]):
        super(MiM, self).__init__()

        # self.deconv3 = PatchExpand2D(channels[4]//2)
        #self.deconv3 = nn.ConvTranspose2d(channels[4], channels[3], 4, 2, 1)
        self.deconv3 = PatchExpand2D(channels[4])

        self.uplayer3 = self._make_layer(block=ResidualBlock, block_num=layer_blocks[2],
                                         in_channels=channels[4], out_channels=channels[4], stride=1)
        self.deconv2 = PatchExpand2D(channels[3])
        #self.deconv2 = nn.ConvTranspose2d(channels[3], channels[2], 4, 2, 1)
        self.uplayer2 = self._make_layer(block=ResidualBlock, block_num=layer_blocks[1],
                                         in_channels=channels[3], out_channels=channels[3], stride=1)
        self.deconv1 = PatchExpand2D(channels[2])
        #self.deconv1 = nn.ConvTranspose2d(channels[2], channels[1], 4, 2, 1)
        self.uplayer1 = self._make_layer(block=ResidualBlock, block_num=layer_blocks[0],
                                         in_channels=channels[2], out_channels=channels[2], stride=1)
        self.head = _FCNHead(channels[2], 1)
        #####################
        self.mim_backbone = PyramidMiM_enc()

    def forward(self, x): # the input is of size (b,3,512,512), the output is of size (b,1,512,512), where the num_class=1 in ISTD.
        _, _, hei, wid = x.shape
        
        outputs=self.mim_backbone(x)
        t1,t2,t3,t4=outputs[0],outputs[1],outputs[2],outputs[3]
        # print('t1 shape', t1.shape)
        # print('t2 shape', t2.shape)
        # print('t3 shape', t3.shape)
        # print('t4 shape', t4.shape)

        deconc3 = self.deconv3(t4)
        fusec3 = deconc3+t3

        # print('fusec3 shape', fusec3.shape)
       
        upc3 = self.uplayer3(fusec3)

        deconc2 = self.deconv2(upc3)
        fusec2 = deconc2+t2

        upc2 = self.uplayer2(fusec2)

        deconc1 = self.deconv1(upc2)
        fusec1 = deconc1+t1
    
        upc1 = self.uplayer1(fusec1)

        # print('upc1 shape', upc1.shape)

        pred = self.head(upc1)
        out = F.interpolate(pred, size=[hei, wid], mode='bilinear')

        return out

    def _make_layer(self, block, block_num, in_channels, out_channels, stride):
        layer = []
        downsample = (in_channels != out_channels) or (stride != 1)
        layer.append(block(in_channels, out_channels, stride, downsample))
        for _ in range(block_num-1):
            layer.append(block(out_channels, out_channels, 1, False))
        return nn.Sequential(*layer)


if __name__ == '__main__':
    input_ = torch.Tensor(5, 1, 256, 256).cuda()
    net = MiM([2]*3,[8, 16, 32, 64, 128]).cuda()
    out=net(input_)

    flops, params = profile(net, (input_,))

    print("-" * 50)
    print('FLOPs = ' + str(flops / 1000 ** 3) + ' G')
    print('Params = ' + str(params / 1000 ** 2) + ' M')






# examples 模块
# 功能：包含使用custom_deep_learning包进行深度学习任务的示例代码。
# 子模块/文件：
# image_classification.py：使用CNN进行图像分类的示例。
# text_generation.py：使用Transformer进行文本生成的示例。
# custom_task.py：用户自定义任务的示例。











# examples/__init__.py

# 这个文件是 examples 包的初始化文件。
# 它主要用于组织和文档化 examples 包中的示例脚本。
# 通常，这里不会定义任何函数或类。

# 如果需要，可以在这里添加一些描述或文档字符串。
"""
This package contains example scripts demonstrating various functionalities
of the custom_deep_learning package.

Available examples:
- mnist_example.py: A script showing how to train a model on the MNIST dataset.
- cifar10_example.py: An example of training a model on the CIFAR-10 dataset.
- custom_data_example.py: A script for training a model on a custom dataset.

To run an example, use the command line:
python -m custom_deep_learning.examples.mnist_example
"""

# 如果需要，也可以在这里导入其他模块或包，但通常这不是必需的。

# 注意：通常情况下，这个文件中不会有 __all__ 变量的定义，
# 因为我们并不希望从 examples 包中直接导入任何对象。


"""
datasets 模块
功能：包含数据集的加载、预处理和增强功能。
子模块/文件：
image_datasets.py：图像数据集的处理，如CIFAR、ImageNet等。
text_datasets.py：文本数据集的处理，如IMDB、WikiText等。
custom_dataset.py：用户自定义数据集的示例或模板。
dataset_utils.py:数据集工具模块，主要定义数据集处理工具.
"""


# 如果需要，可以直接导入这些模块中的特定类或函数
from .image_datasets import load_data_minist, load_data_fashion_mnist

from .text_datasets import read_time_machine, load_corpus_time_machine, SeqDataLoader_time_machine, load_data_time_machine, read_data_nmt, preprocess_nmt, tokenize_nmt, load_data_nmt

from .custom_dataset import load_arrays

from .dataset_utils import download, tokenize, Vocab, seq_data_iter_random, seq_data_iter_sequential, download_extract, truncate_pad, sequence_mask, masked_softmax, get_all_images, get_img_norm_cfg
# from .cifar10_dataset import load_cifar10_data
# from .custom_dataset import load_custom_data
#
# __all__ 变量定义了当使用 from datasets import * 时导入哪些对象
# 注意：通常不推荐使用 from package import *
__all__ = [
    'load_data_minist',
    'load_data_fashion_mnist',
    'read_time_machine',
    'load_corpus_time_machine',
    'SeqDataLoader_time_machine',
    'load_data_time_machine',
    'read_data_nmt',
    'preprocess_nmt',
    'tokenize_nmt',
    'load_data_nmt',
    'load_arrays',
    'download',
    'tokenize',
    'Vocab',
    'seq_data_iter_random',
    'seq_data_iter_sequential',
    'download_extract',
    'truncate_pad',
    'sequence_mask','masked_softmax','get_all_images', 'get_img_norm_cfg'
]





"""
core 模块
功能：包含包的核心功能和基础类。
子模块/文件：
base_model.py：定义基础模型类，提供模型初始化、保存、加载等基础功能。
tools.py：包含一些通用的工具函数，如数据处理、模型评估等。
text.tools: 负责处理文本数据相关的功能，例如词表构建、词频统计等。
"""
# core/__init__.py

from .utils import evaluate_accuracy_gpu
from .utils import train_custom
from .utils import try_gpu
from .utils import try_all_gpus
from .utils import use_svg_display
from .utils import set_figsize
from .utils import set_axes
from .utils import plot
from .utils import evaluate_loss
from .utils import Accumulator
from .utils import Timer, Animator, show_heatmaps

from .image_utils import visualize_grid_cells
from .image_utils import show_images
from .image_utils import show_boxes
from .image_utils import bbox_to_rect
from .image_utils import box_corner_to_center
from .image_utils import visualize_bounding_boxes
from .image_utils import box_iou_xyxy
from .image_utils import box_iou_xywh
from .image_utils import draw_anchor_boxes
from .image_utils import get_objectness_label
from .image_utils import nms
from .image_utils import multiclass_nms
from .image_utils import get_yolo_box_xxyy
from .image_utils import image_downsampling, image_quantization, image_Brightness, image_contrast, image_cv2plt

from .text_utils import show_list_len_pair_hist, bleu

from .digital_image_process import geometric_mean_filter

from . import transforms
__all__ = [
    'evaluate_accuracy_gpu',
    'train_custom',
    'try_gpu',
    'try_all_gpus',
    'set_figsize',
    'set_axes',
    'plot',
    'evaluate_loss',
    'Accumulator',
    'Timer',
    'Animator', 'show_heatmaps',


    'visualize_grid_cells',
    'show_images',
    'show_boxes',
    'bbox_to_rect',
    'box_corner_to_center',
    'visualize_bounding_boxes',
    'box_iou_xyxy',
    'box_iou_xywh',
    'draw_anchor_boxes',
    'get_objectness_label',
    'nms',
    'multiclass_nms',
    'get_yolo_box_xxyy', 'image_downsampling','image_quantization','image_Brightness','image_contrast','image_cv2plt',

    'show_list_len_pair_hist',
    'bleu',

    'transforms',
    'geometric_mean_filter'
]


# IRSTD任务的模型复现代码
from .mim_network import MiM
from .SCTransNet import SCTransNet, get_SCTrans_config
from .IRSTDNet import IRSTDNet



__all__ = ['MiM', 'SCTransNet', 'get_SCTrans_config', 'IRSTDNet']


# from ...environment import torch,np,collections
# from ...environment import os
# import torch,collections,os
# import numpy as np

from PepperPepper.environment import torch,np,collections
from PepperPepper.environment import os


class StdConv2d(torch.nn.Conv2d):
    '''卷积权重进行标准化'''
    def forward(self, x):
        w = self.weight
        v, m = torch.var_mean(w, dim=[1, 2, 3], keepdim=True, unbiased=False)
        w = (w - m) / torch.sqrt(v + 1e-5)
        return torch.nn.functional.conv2d(x, w, self.bias, self.stride, self.padding,self.dilation, self.groups)



def conv3x3(cin, cout, stride=1, groups=1, bias=False):
    return StdConv2d(cin, cout, kernel_size=3, stride=stride,
                     padding=1, bias=bias, groups=groups)


def conv1x1(cin, cout, stride=1, bias=False):
    return StdConv2d(cin, cout, kernel_size=1, stride=stride,
                     padding=0, bias=bias)


class PreActBottleneck(torch.nn.Module):
    """
    Pre-activation (v2) bottleneck block.
    """

    def __init__(self, cin, cout=None, cmid=None, stride=1):
        super().__init__()
        cout = cout or cin
        cmid = cmid or cout//4

        self.gn1 = torch.nn.GroupNorm(32, cmid, eps=1e-6)
        self.conv1 = conv1x1(cin, cmid, bias=False)
        self.gn2 = torch.nn.GroupNorm(32, cmid, eps=1e-6)
        self.conv2 = conv3x3(cmid, cmid, stride, bias=False)  # Original code has it on conv1!!
        self.gn3 = torch.nn.GroupNorm(32, cout, eps=1e-6)
        self.conv3 = conv1x1(cmid, cout, bias=False)
        self.relu = torch.nn.ReLU(inplace=True)

        if (stride != 1 or cin != cout):
            # Projection also with pre-activation according to paper.
            self.downsample = conv1x1(cin, cout, stride, bias=False)
            self.gn_proj = torch.nn.GroupNorm(cout, cout)

    def forward(self, x):

        # Residual branch
        residual = x
        if hasattr(self, 'downsample'):
            residual = self.downsample(x)
            residual = self.gn_proj(residual)

        # Unit's branch
        y = self.relu(self.gn1(self.conv1(x)))
        y = self.relu(self.gn2(self.conv2(y)))
        y = self.gn3(self.conv3(y))

        y = self.relu(residual + y)
        return y




class ResNetV2(torch.nn.Module):
    """Implementation of Pre-activation (v2) ResNet mode."""
    def __init__(self, in_channels, block_units=(3, 4, 9), width_factor=1):
        super().__init__()
        self.in_channels = in_channels
        width = int(width_factor * 64)
        self.width = width


        self.root = torch.nn.Sequential(
            collections.OrderedDict([
                ('conv', StdConv2d(self.in_channels, width, kernel_size=7, stride=2, bias=False, padding=3)),
                ('gn', torch.nn.GroupNorm(32, width, eps=1e-6)),
                ('relu', torch.nn.ReLU(inplace=True)),
            ])
        )


        self.body = torch.nn.Sequential(collections.OrderedDict([
            ('block1', torch.nn.Sequential(collections.OrderedDict(
                [('unit1', PreActBottleneck(cin=width, cout=width*4, cmid=width))] +
                [(f'unit{i:d}', PreActBottleneck(cin=width*4, cout=width*4, cmid=width)) for i in range(2, block_units[0] + 1)],
                ))),
            ('block2', torch.nn.Sequential(collections.OrderedDict(
                [('unit1', PreActBottleneck(cin=width*4, cout=width*8, cmid=width*2, stride=2))] +
                [(f'unit{i:d}', PreActBottleneck(cin=width*8, cout=width*8, cmid=width*2)) for i in range(2, block_units[1] + 1)],
                ))),
            ('block3', torch.nn.Sequential(collections.OrderedDict(
                [('unit1', PreActBottleneck(cin=width*8, cout=width*16, cmid=width*4, stride=2))] +
                [(f'unit{i:d}', PreActBottleneck(cin=width*16, cout=width*16, cmid=width*4)) for i in range(2, block_units[2] + 1)],
                ))),
        ]))

    def forward(self, x):
        features = []
        b, c, in_size, _ = x.size()
        x = self.root(x)
        features.append(x)
        x = torch.nn.MaxPool2d(kernel_size=3, stride=2, padding=0)(x)
        for i in range(len(self.body)-1):
            x = self.body[i](x)
            right_size = int(in_size / 4 / (i+1))
            if x.size()[2] != right_size:
                pad = right_size - x.size()[2]
                assert pad < 3 and pad > 0, "x {} should {}".format(x.size(), right_size)
                feat = torch.zeros((b, x.size()[1], right_size, right_size), device=x.device)
                feat[:, :, 0:x.size()[2], 0:x.size()[3]] = x[:]
            else:
                feat = x
            features.append(feat)
        x = self.body[-1](x)
        return x, features[::-1]








from ..environment import torch, math



'''
1.PositionalEncoding(nn.Module)
    位置编码
'''
#@save
class PositionalEncoding(torch.nn.Module):
    """位置编码"""
    def __init__(self, num_hiddens, dropout, max_len=1000):
        super(PositionalEncoding, self).__init__()
        self.dropout = torch.nn.Dropout(dropout)
        # 创建一个足够长的P
        self.P = torch.zeros((1, max_len, num_hiddens))
        X = torch.arange(max_len, dtype=torch.float32).reshape(
            -1, 1) / torch.pow(10000, torch.arange(
            0, num_hiddens, 2, dtype=torch.float32) / num_hiddens)
        self.P[:, :, 0::2] = torch.sin(X)
        self.P[:, :, 1::2] = torch.cos(X)

    def forward(self, X):
        X = X + self.P[:, :X.shape[1], :].to(X.device)
        return self.dropout(X)




from PepperPepper.environment import torch, nn, math, trunc_normal_, profile
from PepperPepper.IRSTD.models import SCTransNet, MiM, get_SCTrans_config



class IRSTDNet(nn.Module):
    def __init__(self, model_name):
        super(IRSTDNet, self).__init__()
        self.model_name = model_name
        self.cal_loss = nn.BCEWithLogitsLoss()
        self.model = None


        if model_name == 'MiM':
            self.model = MiM()
        elif model_name == 'SCTransNet':
            config = get_SCTrans_config()
            self.model = SCTransNet(config, mode='train', deepsuper=True)
        else:
            raise NotImplementedError

        self.apply(self._init_weights)

    def forward(self, img):
        return self.model(img)

    def loss(self, preds, gt_masks):
        # preds = torch.sigmoid(preds)
        # gt_masks = torch.sigmoid(gt_masks)
        if isinstance(preds, list):
            loss_total = 0
            for i in range(len(preds)):
                pred = preds[i]
                gt_mask = gt_masks[i]
                loss = self.cal_loss(pred, gt_mask)
                loss_total = loss_total + loss
            return loss_total / len(preds)

        elif isinstance(preds, tuple):
            loss_total = 0
            for i in range(len(preds)):
                pred = preds[i]
                loss = self.cal_loss(pred, gt_masks)
                loss_total = loss_total + loss
            return loss_total

        else:
            loss = self.cal_loss(preds, gt_masks)
            return loss


    def lossIoU(self, preds, gt_masks):
        if isinstance(preds, list):
            loss_total = 0
            for i in range(len(preds)):
                pred = preds[i]
                gt_mask = gt_masks[i]
                loss = self.softiou(pred, gt_mask)
                loss_total = loss_total + loss
            return loss_total / len(preds)

        elif isinstance(preds, tuple):
            a = []
            for i in range(len(preds)):
                pred = preds[i]
                loss = self.softiou(pred, gt_masks)
                a.append(loss)
            loss_total = a[0] + a[1] + a[2] + a[3] + a[4] + a[5]
            return loss_total

        else:
            loss = self.softiou(preds, gt_masks)
            return loss




    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        if isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)


if __name__ == '__main__':
    # config_vit = get_SCTrans_config()
    # # model = SCTransNet(config_vit, mode='train', deepsuper=True)
    # model = model

    net = IRSTDNet(model_name='SCTransNet')
    inputs = torch.rand(1, 1, 256, 256)
    output = net(inputs)
    flops, params = profile(net, (inputs,))

    print("-" * 50)
    print('FLOPs = ' + str(flops / 1000 ** 3) + ' G')
    print('Params = ' + str(params / 1000 ** 2) + ' M')


from PepperPepper.environment import ml_collections

def get_b16_config():
    """Returns the ViT-B/16 configuration."""
    config = ml_collections.ConfigDict()
    config.patches = ml_collections.ConfigDict({'size': (16, 16)})
    config.image_channels = 3
    config.hidden_size = 768
    config.transformer = ml_collections.ConfigDict()
    config.transformer.mlp_dim = 3072
    config.transformer.num_heads = 12
    config.transformer.num_layers = 12
    config.transformer.attention_dropout_rate = 0.0
    config.transformer.dropout_rate = 0.1

    config.classifier = 'seg'
    config.representation_size = None
    config.resnet_pretrained_path = None
    config.pretrained_path = '../model/vit_checkpoint/imagenet21k/ViT-B_16.npz'
    config.patch_size = 16

    config.decoder_channels = (256, 128, 64, 16)
    config.n_classes = 2
    config.activation = 'softmax'
    return config


def get_testing():
    """Returns a minimal configuration for testing."""
    config = ml_collections.ConfigDict()
    config.patches = ml_collections.ConfigDict({'size': (16, 16)})
    config.hidden_size = 1
    config.transformer = ml_collections.ConfigDict()
    config.transformer.mlp_dim = 1
    config.transformer.num_heads = 1
    config.transformer.num_layers = 1
    config.transformer.attention_dropout_rate = 0.0
    config.transformer.dropout_rate = 0.1
    config.classifier = 'token'
    config.representation_size = None
    return config

def get_r50_b16_config():
    # ResNetV2(block_units=config.resnet.num_layers, width_factor=config.resnet.width_factor)
    """Returns the Resnet50 + ViT-B/16 configuration."""

    config = get_b16_config()
    config.patches.grid = (16, 16)
    config.resnet = ml_collections.ConfigDict()
    config.resnet.num_layers = (3, 4, 9)
    config.resnet.width_factor = 1

    config.classifier = 'seg'
    config.pretrained_path = '../model/vit_checkpoint/imagenet21k/R50+ViT-B_16.npz'
    config.decoder_channels = (256, 128, 64, 16)
    config.skip_channels = [512, 256, 64, 16]
    config.n_classes = 1
    config.n_skip = 3
    config.activation = 'softmax'
    # config.image_channels = 1

    return config


def get_b32_config():
    """Returns the ViT-B/32 configuration."""
    config = get_b16_config()
    config.patches.size = (32, 32)
    config.pretrained_path = '../model/vit_checkpoint/imagenet21k/ViT-B_32.npz'
    return config


def get_l16_config():
    """Returns the ViT-L/16 configuration."""
    config = ml_collections.ConfigDict()
    config.patches = ml_collections.ConfigDict({'size': (16, 16)})
    config.hidden_size = 1024
    config.transformer = ml_collections.ConfigDict()
    config.transformer.mlp_dim = 4096
    config.transformer.num_heads = 16
    config.transformer.num_layers = 24
    config.transformer.attention_dropout_rate = 0.0
    config.transformer.dropout_rate = 0.1
    config.representation_size = None

    # custom
    config.classifier = 'seg'
    config.resnet_pretrained_path = None
    config.pretrained_path = '../model/vit_checkpoint/imagenet21k/ViT-L_16.npz'
    config.decoder_channels = (256, 128, 64, 16)
    config.n_classes = 2
    config.activation = 'softmax'
    return config


def get_r50_l16_config():
    """Returns the Resnet50 + ViT-L/16 configuration. customized """
    config = get_l16_config()
    config.patches.grid = (16, 16)
    config.resnet = ml_collections.ConfigDict()
    config.resnet.num_layers = (3, 4, 9)
    config.resnet.width_factor = 1

    config.classifier = 'seg'
    config.resnet_pretrained_path = '../model/vit_checkpoint/imagenet21k/R50+ViT-B_16.npz'
    config.decoder_channels = (256, 128, 64, 16)
    config.skip_channels = [512, 256, 64, 16]
    config.n_classes = 2
    config.activation = 'softmax'
    return config


def get_l32_config():
    """Returns the ViT-L/32 configuration."""
    config = get_l16_config()
    config.patches.size = (32, 32)
    return config


def get_h14_config():
    """Returns the ViT-L/16 configuration."""
    config = ml_collections.ConfigDict()
    config.patches = ml_collections.ConfigDict({'size': (14, 14)})
    config.hidden_size = 1280
    config.transformer = ml_collections.ConfigDict()
    config.transformer.mlp_dim = 5120
    config.transformer.num_heads = 16
    config.transformer.num_layers = 32
    config.transformer.attention_dropout_rate = 0.0
    config.transformer.dropout_rate = 0.1
    config.classifier = 'token'
    config.representation_size = None

    return config



from PepperPepper.environment import torch
from PepperPepper.models.TransUNet.Embeddings import Embeddings
from PepperPepper.models.TransUNet.Encoder import Encoder

class Transformer(torch.nn.Module):
    def __init__(self, config, img_size, vis):
        super(Transformer, self).__init__()
        self.embeddings = Embeddings(config, img_size=img_size, in_channels=config.image_channels)
        self.encoder = Encoder(config, vis)

    def forward(self, input_ids):
        embedding_output, features = self.embeddings(input_ids)
        encoded, attn_weights = self.encoder(embedding_output)  # (B, n_patch, hidden)
        return encoded, attn_weights, features

